# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import copy
import bz2
import gzip
import io
import logging
import posixpath
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Union, Tuple
from datetime import datetime, date
from decimal import Decimal

import pyarrow as pa
import numpy as np
import pyarrow.compute as pc
import pyarrow.fs as pafs
from pyarrow.parquet import ParquetFile

from fsspec import AbstractFileSystem
from pyarrow import csv as pacsv
from pyarrow import feather as paf
from pyarrow import json as pajson
from pyarrow import parquet as papq
from pyarrow import orc as paorc
from ray.data.datasource import FilenameProvider

from deltacat import logs
from deltacat.exceptions import ContentTypeValidationError
from deltacat.types.media import (
    DELIMITED_TEXT_CONTENT_TYPES,
    TABULAR_CONTENT_TYPES,
    ContentEncoding,
    ContentType,
)
from deltacat.types.partial_download import (
    PartialFileDownloadParams,
    PartialParquetParameters,
)
from deltacat.utils.common import ContentTypeKwargsProvider, ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.schema import coerce_pyarrow_table_to_schema
from deltacat.utils.arguments import (
    sanitize_kwargs_to_callable,
    sanitize_kwargs_by_supported_kwargs,
)
from deltacat.utils.filesystem import resolve_path_and_filesystem
from functools import lru_cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deltacat.storage.model.manifest import Manifest
    from deltacat.storage.model.delta import Delta


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

RAISE_ON_EMPTY_CSV_KWARG = "raise_on_empty_csv"
READER_TYPE_KWARG = "reader_type"
OVERRIDE_CONTENT_ENCODING_FOR_PARQUET_KWARG = "override_content_encoding_for_parquet"

"""
By default, round decimal values using half_to_even round mode when
rescaling a decimal to the given scale and precision in the schema would cause
data loss. Setting any non null value of this argument will result
in an error instead.
"""
RAISE_ON_DECIMAL_OVERFLOW = "raise_on_decimal_overflow"
# Note the maximum from https://arrow.apache.org/docs/python/generated/pyarrow.Decimal256Type.html#pyarrow.Decimal256Type
DECIMAL256_DEFAULT_SCALE = 38
DECIMAL256_MAX_PRECISION = 76
MAX_INT_BYTES = 2147483646


def _filter_schema_for_columns(schema: pa.Schema, columns: List[str]) -> pa.Schema:

    target_schema_fields = []

    for column_name in columns:
        index = schema.get_field_index(column_name)

        if index != -1:
            target_field = schema.field(index)
            target_schema_fields.append(target_field)

    target_schema = pa.schema(target_schema_fields, metadata=schema.metadata)

    return target_schema


def _extract_arrow_schema_from_read_csv_kwargs(kwargs: Dict[str, Any]) -> pa.Schema:
    schema = None
    if (
        "convert_options" in kwargs
        and kwargs["convert_options"].column_types is not None
    ):
        schema = kwargs["convert_options"].column_types
        if not isinstance(schema, pa.Schema):
            schema = pa.schema(schema)
        if kwargs["convert_options"].include_columns:
            schema = _filter_schema_for_columns(
                schema, kwargs["convert_options"].include_columns
            )
        elif (
            kwargs.get("read_options") is not None
            and kwargs["read_options"].column_names
        ):
            schema = _filter_schema_for_columns(
                schema, kwargs["read_options"].column_names
            )
    else:
        logger.debug(
            "Schema not specified in the kwargs."
            " Hence, schema could not be inferred from the empty CSV."
        )

    return schema


def _new_schema_with_replaced_fields(
    schema: pa.Schema, field_to_replace: Callable[[pa.Field], Optional[pa.Field]]
) -> pa.Schema:
    if schema is None:
        return None

    new_schema_fields = []
    for field in schema:
        new_field = field_to_replace(field)
        if new_field is not None:
            new_schema_fields.append(new_field)
        else:
            new_schema_fields.append(field)

    return pa.schema(new_schema_fields, metadata=schema.metadata)


def _read_csv_rounding_decimal_columns_to_fit_scale(
    schema: pa.Schema, reader_args: List[Any], reader_kwargs: Dict[str, Any]
) -> pa.Table:
    # Note: We read decimals as strings first because CSV
    # conversion to decimal256 isn't implemented as of pyarrow==12.0.1
    new_schema = _new_schema_with_replaced_fields(
        schema,
        lambda fld: (
            pa.field(fld.name, pa.string(), metadata=fld.metadata)
            if pa.types.is_decimal128(fld.type) or pa.types.is_decimal256(fld.type)
            else None
        ),
    )
    new_kwargs = sanitize_kwargs_by_supported_kwargs(
        ["read_options", "parse_options", "convert_options", "memory_pool"],
        reader_kwargs,
    )
    # Creating a shallow copy for efficiency
    new_convert_options = copy.copy(new_kwargs["convert_options"])
    new_convert_options.column_types = new_schema
    new_reader_kwargs = {**new_kwargs, "convert_options": new_convert_options}
    arrow_table = pacsv.read_csv(*reader_args, **new_reader_kwargs)

    for column_index, field in enumerate(schema):
        if pa.types.is_decimal128(field.type) or pa.types.is_decimal256(field.type):
            column_array = arrow_table[field.name]
            # We always cast to decimal256 to accomodate fixed scale of 38
            cast_to_type = pa.decimal256(
                DECIMAL256_MAX_PRECISION, DECIMAL256_DEFAULT_SCALE
            )
            casted_decimal_array = pc.cast(column_array, cast_to_type)
            # Note that scale can be negative
            rounded_column_array = pc.round(
                casted_decimal_array, ndigits=field.type.scale
            )
            final_decimal_array = pc.cast(rounded_column_array, field.type)
            arrow_table = arrow_table.set_column(
                column_index,
                field,
                final_decimal_array,
            )
            logger.debug(
                f"Rounded decimal column: {field.name} to {field.type.scale} scale and"
                f" {field.type.precision} precision"
            )

    return arrow_table


def pyarrow_read_csv_default(*args, **kwargs):
    new_kwargs = sanitize_kwargs_by_supported_kwargs(
        ["read_options", "parse_options", "convert_options", "memory_pool"], kwargs
    )

    try:
        return pacsv.read_csv(*args, **new_kwargs)
    except pa.lib.ArrowInvalid as e:
        error_str = e.__str__()
        schema = _extract_arrow_schema_from_read_csv_kwargs(kwargs)

        if error_str == "Empty CSV file" and not kwargs.get(RAISE_ON_EMPTY_CSV_KWARG):
            logger.debug(f"Read CSV empty schema being used: {schema}")
            return pa.Table.from_pylist([], schema=schema)
        if not kwargs.get(RAISE_ON_DECIMAL_OVERFLOW):
            # Note, this logic requires expensive casting. To prevent downgrading performance
            # for happy path reads, we are handling this case in response to an error.
            logger.warning(
                "Rescaling Decimal to the given scale in the schema. "
                f"Original error: {error_str}"
            )

            if schema is not None and "convert_options" in kwargs:
                if (
                    "Rescaling Decimal" in error_str
                    and "value would cause data loss" in error_str
                ):
                    logger.debug(f"Checking if the file: {args[0]}...")
                    # Since we are re-reading the file, we have to seek to beginning
                    if isinstance(args[0], io.IOBase) and args[0].seekable():
                        logger.debug(f"Seeking to the beginning of the file {args[0]}")
                        args[0].seek(0)
                    return _read_csv_rounding_decimal_columns_to_fit_scale(
                        schema=schema, reader_args=args, reader_kwargs=kwargs
                    )
            else:
                logger.debug(
                    "Schema is None when trying to adjust decimal values. "
                    "Hence, bubbling up exception..."
                )

        raise e


# TODO(pdames): Remove deprecated S3-only readers.
def read_csv(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Filter out DeltaCAT-specific parameters that PyArrow doesn't understand
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    pyarrow_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    # TODO(pdames): Merge in decimal256 support from pure S3 path reader.

    # Check if compression is already indicated by file path
    should_decompress = path.endswith(".gz")

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
            # Handle decompression - avoid double decompression for PyArrow filesystem
            if should_decompress:
                # PyArrow filesystem already handles .gz decompression automatically
                return pacsv.read_csv(f, **pyarrow_kwargs)
            else:
                # Apply explicit decompression if needed
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    return pacsv.read_csv(input_file, **pyarrow_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle decompression - apply explicit decompression for fsspec
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return pacsv.read_csv(input_file, **pyarrow_kwargs)


def read_feather(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Filter out DeltaCAT-specific parameters that PyArrow doesn't understand
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    pyarrow_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return paf.read_table(input_file, **pyarrow_kwargs)
    else:
        # fsspec AbstractFileSystem - Feather requires seekable files
        # For local files, we can use the file path directly
        if hasattr(filesystem, "protocol") and filesystem.protocol == "file":
            if content_encoding != ContentEncoding.IDENTITY.value:
                # For compressed files, decompress to a temporary file
                import tempfile
                import shutil

                with filesystem.open(path, "rb", **fs_open_kwargs) as f:
                    input_file_init = ENCODING_TO_FILE_INIT.get(
                        content_encoding, lambda x: x
                    )
                    with input_file_init(f) as input_file:
                        # Create temporary file to hold decompressed data
                        with tempfile.NamedTemporaryFile() as temp_file:
                            shutil.copyfileobj(input_file, temp_file)
                            temp_file.flush()
                            return paf.read_table(temp_file.name, **read_kwargs)
            else:
                # No compression, can read directly from file path
                return paf.read_table(path, **pyarrow_kwargs)
        else:
            # For non-local filesystems, always read from temporary file
            import tempfile
            import shutil

            with filesystem.open(path, "rb", **fs_open_kwargs) as f:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Create temporary file to hold data
                    with tempfile.NamedTemporaryFile() as temp_file:
                        shutil.copyfileobj(input_file, temp_file)
                        temp_file.flush()
                        return paf.read_table(temp_file.name, **read_kwargs)


def read_json(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Filter out DeltaCAT-specific parameters that PyArrow doesn't understand
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    pyarrow_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    # Check if decompression is already indicated by file path
    should_decompress = path.endswith(".gz")

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
            # Handle decompression - avoid double decompression for PyArrow filesystem
            if should_decompress:
                # PyArrow filesystem already handles .gz decompression automatically
                return pajson.read_json(f, **pyarrow_kwargs)
            else:
                # Apply explicit decompression if needed
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    return pajson.read_json(input_file, **pyarrow_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle decompression - apply explicit decompression for fsspec
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return pajson.read_json(input_file, **pyarrow_kwargs)


def read_orc(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Filter out DeltaCAT-specific parameters that PyArrow doesn't understand
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    pyarrow_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return paorc.read_table(input_file, **pyarrow_kwargs)
    else:
        # fsspec AbstractFileSystem - ORC requires seekable files, so handle compression differently
        if content_encoding != ContentEncoding.IDENTITY.value:
            # For compressed files with fsspec, we need to decompress to a temporary file
            # since ORC requires seekable streams
            import tempfile
            import shutil

            with filesystem.open(path, "rb", **fs_open_kwargs) as f:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Create temporary file to hold decompressed data
                    with tempfile.NamedTemporaryFile() as temp_file:
                        shutil.copyfileobj(input_file, temp_file)
                        temp_file.flush()
                        return paorc.read_table(temp_file.name, **pyarrow_kwargs)
        else:
            # No compression, can read directly
            with filesystem.open(path, "rb", **fs_open_kwargs) as f:
                return paorc.read_table(f, **pyarrow_kwargs)


def read_parquet(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Convert DeltaCAT Schema to PyArrow Schema if present
    if "schema" in read_kwargs:
        from deltacat.storage.model.schema import Schema as DeltaCATSchema

        schema = read_kwargs["schema"]
        if isinstance(schema, DeltaCATSchema):
            read_kwargs["schema"] = schema.arrow

    # Filter out DeltaCAT-specific parameters that PyArrow doesn't understand
    # Use local import to avoid circular dependency
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    pyarrow_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return papq.read_table(input_file, **pyarrow_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                return papq.read_table(input_file, **pyarrow_kwargs)


def read_avro(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pa.Table:
    # Filter out DeltaCAT-specific parameters that Polars doesn't understand
    from deltacat.types.tables import _filter_kwargs_for_external_readers

    polars_kwargs = _filter_kwargs_for_external_readers(read_kwargs)
    """
    Read an Avro file using polars and convert to PyArrow.
    """
    import polars as pl

    # If path is a file-like object, read directly
    if hasattr(path, "read"):
        pl_df = pl.read_avro(path, **polars_kwargs)
        return pl_df.to_arrow()

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
            # Handle compression
            input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
            with input_file_init(f) as input_file:
                pl_df = pl.read_avro(input_file, **polars_kwargs)
                return pl_df.to_arrow()
    with filesystem.open(path, "rb", **fs_open_kwargs) as f:
        input_file_init = ENCODING_TO_FILE_INIT.get(content_encoding, lambda x: x)
        with input_file_init(f) as input_file:
            pl_df = pl.read_avro(input_file, **polars_kwargs)
            return pl_df.to_arrow()


def pyarrow_read_csv(*args, **kwargs) -> pa.Table:
    schema = _extract_arrow_schema_from_read_csv_kwargs(kwargs)

    # CSV conversion to decimal256 isn't supported as of pyarrow=12.0.1
    # Below ensures decimal256 is casted properly.
    schema_includes_decimal256 = (
        (True if any([pa.types.is_decimal256(x.type) for x in schema]) else False)
        if schema is not None
        else None
    )
    if schema_includes_decimal256 and not kwargs.get(RAISE_ON_DECIMAL_OVERFLOW):
        # falling back to expensive method of reading CSV
        return _read_csv_rounding_decimal_columns_to_fit_scale(
            schema, reader_args=args, reader_kwargs=kwargs
        )
    else:
        return pyarrow_read_csv_default(*args, **kwargs)


CONTENT_TYPE_TO_PA_S3_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pyarrow_read_csv,
    ContentType.TSV.value: pyarrow_read_csv,
    ContentType.CSV.value: pyarrow_read_csv,
    ContentType.PSV.value: pyarrow_read_csv,
    ContentType.PARQUET.value: papq.read_table,
    ContentType.FEATHER.value: paf.read_table,
    ContentType.JSON.value: pajson.read_json,
    ContentType.ORC.value: paorc.read_table,
    ContentType.AVRO.value: read_avro,
}


CONTENT_TYPE_TO_READ_FN: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: read_csv,
    ContentType.TSV.value: read_csv,
    ContentType.CSV.value: read_csv,
    ContentType.PSV.value: read_csv,
    ContentType.PARQUET.value: read_parquet,
    ContentType.FEATHER.value: read_feather,
    ContentType.JSON.value: read_json,
    ContentType.ORC.value: read_orc,
    ContentType.AVRO.value: read_avro,
}


def write_feather(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            paf.write_feather(table, f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            paf.write_feather(table, f, **write_kwargs)


def write_csv(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if write_kwargs.get("write_options") is None:
        # column names are kept in table metadata, so omit header
        write_kwargs["write_options"] = pacsv.WriteOptions(include_header=False)

    # Check if the path already indicates compression to avoid double compression
    should_compress = path.endswith(".gz")

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            if should_compress:
                # Path ends with .gz, PyArrow filesystem automatically compresses, no need for additional compression
                pacsv.write_csv(table, f, **write_kwargs)
            else:
                # No compression indicated, write uncompressed
                pacsv.write_csv(table, f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            if should_compress:
                # For fsspec filesystems, we need to apply compression explicitly
                with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
                    pacsv.write_csv(table, out, **write_kwargs)
            else:
                # No compression indicated, write uncompressed
                pacsv.write_csv(table, f, **write_kwargs)


def write_orc(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            paorc.write_table(table, f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            paorc.write_table(table, f, **write_kwargs)


def write_parquet(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            papq.write_table(table, f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            papq.write_table(table, f, **write_kwargs)


def write_json(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    """
    Write a PyArrow Table to a JSON file by delegating to polars implementation.
    """
    import polars as pl
    from deltacat.utils.polars import write_json as polars_write_json

    # Convert PyArrow Table to polars DataFrame
    pl_df = pl.from_arrow(table)

    # Delegate to polars write_json implementation with GZIP compression
    polars_write_json(
        pl_df,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        **write_kwargs,
    )


def write_avro(
    table: pa.Table,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    """
    Write a PyArrow Table to an AVRO file by delegating to polars implementation.
    """
    import polars as pl
    from deltacat.utils.polars import write_avro as polars_write_avro

    # Convert PyArrow Table to polars DataFrame
    pl_df = pl.from_arrow(table)

    # Delegate to polars write_avro implementation
    polars_write_avro(
        pl_df,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        **write_kwargs,
    )


CONTENT_TYPE_TO_PA_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: write_csv,
    ContentType.TSV.value: write_csv,
    ContentType.CSV.value: write_csv,
    ContentType.PSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
    ContentType.FEATHER.value: write_feather,
    ContentType.JSON.value: write_json,
    ContentType.AVRO.value: write_avro,
    ContentType.ORC.value: write_orc,
}


def content_type_to_writer_kwargs(content_type: str) -> Dict[str, Any]:
    """
    Returns writer kwargs for the given content type when writing with pyarrow.
    """
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "write_options": pacsv.WriteOptions(
                delimiter="\t",
                include_header=False,
                quoting_style="none",
            )
        }
    if content_type == ContentType.TSV.value:
        return {
            "write_options": pacsv.WriteOptions(
                include_header=False,
                delimiter="\t",
                quoting_style="needed",
            )
        }
    if content_type == ContentType.CSV.value:
        return {
            "write_options": pacsv.WriteOptions(
                include_header=False,
                delimiter=",",
                quoting_style="needed",
            )
        }
    if content_type == ContentType.PSV.value:
        return {
            "write_options": pacsv.WriteOptions(
                include_header=False,
                delimiter="|",
                quoting_style="needed",
            )
        }
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.JSON.value,
        ContentType.AVRO.value,
        ContentType.ORC.value,
    }:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "parse_options": pacsv.ParseOptions(delimiter="\t", quote_char=False),
            "convert_options": pacsv.ConvertOptions(
                null_values=[""],  # pyarrow defaults are ["", "NULL", "null"]
                strings_can_be_null=True,
            ),
        }
    if content_type == ContentType.TSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="\t")}
    if content_type == ContentType.CSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter=",")}
    if content_type == ContentType.PSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="|")}
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.JSON.value,
        ContentType.ORC.value,
        ContentType.AVRO.value,
    }:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


# TODO (pdames): add deflate and snappy
ENCODING_TO_FILE_INIT: Dict[str, Callable] = {
    ContentEncoding.GZIP.value: partial(gzip.open, mode="rb"),
    ContentEncoding.BZIP2.value: partial(bz2.open, mode="rb"),
    ContentEncoding.IDENTITY.value: lambda file_path: file_path,
}


def slice_table(
    table: pa.Table,
    max_len: Optional[int],
) -> List[pa.Table]:
    """
    Iteratively create 0-copy table slices.
    """
    if max_len is None:
        return [table]
    tables = []
    offset = 0
    records_remaining = len(table)
    while records_remaining > 0:
        records_this_entry = min(max_len, records_remaining)
        tables.append(table.slice(offset, records_this_entry))
        records_remaining -= records_this_entry
        offset += records_this_entry
    return tables


def append_column_to_table(
    table: pa.Table,
    column_name: str,
    column_value: Any,
) -> pa.Table:
    num_rows = table.num_rows
    column_array = pa.array([column_value] * num_rows)
    return table.append_column(column_name, column_array)


def select_columns(
    table: pa.Table,
    column_names: List[str],
) -> pa.Table:
    return table.select(column_names)


class ReadKwargsProviderPyArrowCsvPureUtf8(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""

    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options: pacsv.ConvertOptions = kwargs.get("convert_options")
            if convert_options is None:
                convert_options = pacsv.ConvertOptions()
            # read only the included columns as strings?
            column_names = (
                self.include_columns
                if self.include_columns
                else convert_options.include_columns
            )
            if not column_names:
                # read all columns as strings?
                read_options: pacsv.ReadOptions = kwargs.get("read_options")
                if read_options and read_options.column_names:
                    column_names = read_options.column_names
                else:
                    raise ValueError("No column names found!")
            convert_options.column_types = {
                column_name: pa.string() for column_name in column_names
            }
            kwargs["convert_options"] = convert_options
        return kwargs


class ReadKwargsProviderPyArrowSchemaOverride(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that explicitly maps column names to column types when
    loading dataset files into a PyArrow table. Disables the default type inference
    behavior on the defined columns."""

    def __init__(
        self,
        schema: Optional[pa.Schema] = None,
        pq_coerce_int96_timestamp_unit: Optional[str] = None,
        parquet_reader_type: Optional[str] = None,
        file_read_timeout_ms: Optional[int] = None,
    ):
        """

        Args:
            schema: The schema to use for reading the dataset.
                If unspecified, the schema will be inferred from the source.
            pq_coerce_int96_timestamp_unit: When reading from parquet files, cast timestamps that are stored in INT96
                format to a particular resolution (e.g. 'ms').

        """
        self.schema = schema
        self.pq_coerce_int96_timestamp_unit = pq_coerce_int96_timestamp_unit
        self.parquet_reader_type = parquet_reader_type
        self.file_read_timeout_ms = file_read_timeout_ms

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options = kwargs.get("convert_options", pacsv.ConvertOptions())
            if self.schema:
                convert_options.column_types = self.schema
            kwargs["convert_options"] = convert_options
        elif content_type == ContentType.PARQUET:
            # kwargs here are passed into `pyarrow.parquet.read_table`.
            # Only supported in PyArrow 8.0.0+
            if self.schema:
                kwargs["schema"] = self.schema

            # Coerce deprecated int96 timestamp to millisecond if unspecified
            if self.pq_coerce_int96_timestamp_unit is not None:
                kwargs[
                    "coerce_int96_timestamp_unit"
                ] = self.pq_coerce_int96_timestamp_unit

            if self.parquet_reader_type:
                kwargs["reader_type"] = self.parquet_reader_type
            else:
                kwargs["reader_type"] = "daft"

            kwargs["file_timeout_ms"] = self.file_read_timeout_ms

        return kwargs


def _add_column_kwargs(
    content_type: str,
    column_names: Optional[List[str]],
    include_columns: Optional[List[str]],
    kwargs: Dict[str, Any],
):

    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        read_options: pacsv.ReadOptions = kwargs.get("read_options")
        if read_options is None:
            read_options = pacsv.ReadOptions()
        if column_names:
            read_options.column_names = column_names
        else:
            read_options.autogenerate_column_names = True
        kwargs["read_options"] = read_options
        convert_options: pacsv.ConvertOptions = kwargs.get("convert_options")
        if convert_options is None:
            convert_options = pacsv.ConvertOptions()
        if include_columns:
            convert_options.include_columns = include_columns
        kwargs["convert_options"] = convert_options
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            kwargs["columns"] = include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}"
                )


def partial_parquet_file_to_table(
    path: str,
    content_type: str,
    content_encoding: str,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialParquetParameters] = None,
    **kwargs,
) -> pa.Table:

    assert (
        partial_file_download_params is not None
    ), "Partial parquet params must not be None"
    assert (
        partial_file_download_params.row_groups_to_download is not None
    ), "No row groups to download"

    # Resolve filesystem and path
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)

    pq_file = file_to_parquet(
        path=path,
        content_type=content_type,
        content_encoding=content_encoding,
        filesystem=filesystem,
        partial_file_download_params=partial_file_download_params,
        pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        **kwargs,
    )

    table, latency = timed_invocation(
        pq_file.read_row_groups,
        partial_file_download_params.row_groups_to_download,
        columns=include_columns or column_names,
    )

    logger.debug(f"Successfully read from path={path} in {latency}s")

    kwargs = {}

    if pa_read_func_kwargs_provider:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    # Note: ordering is consistent with the `input_schema` if provided
    if kwargs.get("schema") is not None:
        input_schema = kwargs.get("schema")
        if include_columns is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in include_columns],
                metadata=input_schema.metadata,
            )
        elif column_names is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in column_names],
                metadata=input_schema.metadata,
            )
        coerced_table, coerce_latency = timed_invocation(
            coerce_pyarrow_table_to_schema, table, input_schema
        )

        logger.debug(
            f"Coercing the PyArrow table of len {len(coerced_table)} "
            f"into passed schema took {coerce_latency}s"
        )

        return coerced_table

    return table


def table_size(table: pa.Table) -> int:
    return table.nbytes


def parquet_file_size(table: papq.ParquetFile) -> int:
    return table.metadata.serialized_size


def table_to_file(
    table: pa.Table,
    base_path: str,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    content_type: str = ContentType.PARQUET.value,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> None:
    """
    Writes the given Pyarrow Table to a file.

    Args:
        table: The PyArrow table to write
        base_path: Base path to write to
        file_system: Optional filesystem to use
        block_path_provider: Provider for block path generation
        content_type: Content type for the output file
        schema: Optional schema (for compatibility with explicit schema parameter pattern)
        kwargs: Keyword arguments passed to the PyArrow write function
    """
    writer = CONTENT_TYPE_TO_PA_WRITE_FUNC.get(content_type)
    if not writer:
        raise NotImplementedError(
            f"Pyarrow writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PA_WRITE_FUNC.keys}"
        )
    filename = block_path_provider(base_path)
    path = posixpath.join(base_path, filename)
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    logger.debug(f"Writing table: {table} with kwargs: {writer_kwargs} to path: {path}")
    writer(table, path, filesystem=filesystem, **writer_kwargs)


class RecordBatchTables:
    def __init__(self, batch_size: int):
        """
        Data structure for maintaining a batched list of tables, where each batched table has
        a record count of some multiple of the specified record batch size.

        Remaining records are stored in a separate list of tables.

        Args:
            batch_size: Minimum record count per table to batch by. Batched tables are
             guaranteed to have a record count multiple of the batch_size.
        """
        self._batched_tables: List[pa.Table] = []
        self._batched_record_count: int = 0
        self._remaining_tables: List[pa.Table] = []
        self._remaining_record_count: int = 0
        self._batch_size: int = batch_size

    def append(self, table: pa.Table) -> None:
        """
        Appends a table for batching.

        Table record counts are added to any previous remaining record count.
        If the new remainder record count meets or exceeds the configured batch size record count,
        the remainder will be shifted over to the list of batched tables in FIFO order via table slicing.
        Batched tables will always have a record count of some multiple of the configured batch size.

        Record ordering is preserved from input tables whenever tables are shifted from the remainder
        over to the batched list. Records from Table A will always precede records from Table B,
        if Table A was appended before Table B. Records from the batched list will always precede records
        from the remainders.

        Ex:
            bt = RecordBatchTables(8)
            col1 = pa.array([i for i in range(10)])
            test_table = pa.Table.from_arrays([col1], names=["col1"])
            bt.append(test_table)

            print(bt.batched_records)  # 8
            print(bt.batched)  # [0, 1, 2, 3, 4, 5, 6, 7]
            print(bt.remaining_records)  # 2
            print(bt.remaining)  # [8, 9]

        Args:
            table: Input table to add

        """
        if self._remaining_tables:
            if self._remaining_record_count + len(table) < self._batch_size:
                self._remaining_tables.append(table)
                self._remaining_record_count += len(table)
                return

            records_to_fit = self._batch_size - self._remaining_record_count
            fitted_table = table.slice(length=records_to_fit)
            self._remaining_tables.append(fitted_table)
            self._remaining_record_count += len(fitted_table)
            table = table.slice(offset=records_to_fit)

        record_count = len(table)
        record_multiplier, records_leftover = (
            record_count // self._batch_size,
            record_count % self._batch_size,
        )

        if record_multiplier > 0:
            batched_table = table.slice(length=record_multiplier * self._batch_size)
            # Add to remainder tables to preserve record ordering
            self._remaining_tables.append(batched_table)
            self._remaining_record_count += len(batched_table)

        if self._remaining_tables:
            self._shift_remaining_to_new_batch()

        if records_leftover > 0:
            leftover_table = table.slice(offset=record_multiplier * self._batch_size)
            self._remaining_tables.append(leftover_table)
            self._remaining_record_count += len(leftover_table)

    def _shift_remaining_to_new_batch(self) -> None:
        new_batch = pa.concat_tables(self._remaining_tables)
        self._batched_tables.append(new_batch)
        self._batched_record_count += self._remaining_record_count
        self.clear_remaining()

    @staticmethod
    def from_tables(tables: List[pa.Table], batch_size: int) -> RecordBatchTables:
        """
        Static factory for generating batched tables and remainders given a list of input tables.

        Args:
            tables: A list of input tables with various record counts
            batch_size: Minimum record count per table to batch by. Batched tables are
             guaranteed to have a record count multiple of the batch_size.

        Returns: A batched tables object

        """
        rbt = RecordBatchTables(batch_size)
        for table in tables:
            rbt.append(table)
        return rbt

    @property
    def batched(self) -> List[pa.Table]:
        """
        List of tables batched and ready for processing.
        Each table has N records, where N records are some multiple of the configured records batch size.

        For example, if the configured batch size is 5, then a list of batched tables
        could have the following record counts: [60, 5, 30, 10]

        Returns: a list of batched tables

        """
        return self._batched_tables

    @property
    def batched_record_count(self) -> int:
        """
        The number of total records from the batched list.

        Returns: batched record count

        """
        return self._batched_record_count

    @property
    def remaining(self) -> List[pa.Table]:
        """
        List of tables carried over from table slicing during the batching operation.
        The sum of all record counts in the remaining tables is guaranteed to be less than the configured batch size.

        Returns: a list of remaining tables

        """
        return self._remaining_tables

    @property
    def remaining_record_count(self) -> int:
        """
        The number of total records from the remaining tables list.

        Returns: remaining record count

        """
        return self._remaining_record_count

    @property
    def batch_size(self) -> int:
        """
        The configured batch size.

        Returns: batch size

        """
        return self._batch_size

    def has_batches(self) -> bool:
        """
        Checks if there are any currently batched tables ready for processing.

        Returns: true if batched records exist, otherwise false

        """
        return self._batched_record_count > 0

    def has_remaining(self) -> bool:
        """
        Checks if any remaining tables exist after batching.

        Returns: true if remaining records exist, otherwise false

        """
        return self._remaining_record_count > 0

    def evict(self) -> List[pa.Table]:
        """
        Evicts all batched tables from this object and returns them.

        Returns: a list of batched tables

        """
        evicted_tables = [*self.batched]
        self.clear_batches()
        return evicted_tables

    def clear_batches(self) -> None:
        """
        Removes all batched tables and resets batched records.

        """
        self._batched_tables.clear()
        self._batched_record_count = 0

    def clear_remaining(self) -> None:
        """
        Removes all remaining tables and resets remaining records.

        """
        self._remaining_tables.clear()
        self._remaining_record_count = 0


@lru_cache(maxsize=1)
def _int_max_string_len() -> int:
    PA_UINT64_MAX_STR_BYTES = pc.binary_length(
        pc.cast(pa.scalar(2**64 - 1, type=pa.uint64()), pa.string())
    ).as_py()
    PA_INT64_MAX_STR_BYTES = pc.binary_length(
        pc.cast(pa.scalar(-(2**63), type=pa.int64()), pa.string())
    ).as_py()
    return max(PA_UINT64_MAX_STR_BYTES, PA_INT64_MAX_STR_BYTES)


@lru_cache(maxsize=1)
def _float_max_string_len() -> int:
    PA_POS_FLOAT64_MAX_STR_BYTES = pc.binary_length(
        pc.cast(pa.scalar(np.finfo(np.float64).max, type=pa.float64()), pa.string())
    ).as_py()
    PA_NEG_FLOAT64_MAX_STR_BYTES = pc.binary_length(
        pc.cast(pa.scalar(np.finfo(np.float64).min, type=pa.float64()), pa.string())
    ).as_py()
    return max(PA_POS_FLOAT64_MAX_STR_BYTES, PA_NEG_FLOAT64_MAX_STR_BYTES)


def _max_decimal128_string_len():
    return 40  # "-" + 38 digits + decimal


def _max_decimal256_string_len():
    return 78  # "-" + 76 digits + decimal


def sliced_string_cast(array: pa.ChunkedArray) -> pa.ChunkedArray:
    """performs slicing of a pyarrow array prior casting to a string.
    This prevents a pyarrow from allocating too large of an array causing a failure.
    Issue: https://github.com/apache/arrow/issues/38835
    TODO: deprecate this function when pyarrow performs proper ChunkedArray -> ChunkedArray casting
    """
    dtype = array.type
    max_str_len = None
    if pa.types.is_integer(dtype):
        max_str_len = _int_max_string_len()
    elif pa.types.is_floating(dtype):
        max_str_len = _float_max_string_len()
    elif pa.types.is_decimal128(dtype):
        max_str_len = _max_decimal128_string_len()
    elif pa.types.is_decimal256(dtype):
        max_str_len = _max_decimal256_string_len()

    if max_str_len is not None:
        max_elems_per_chunk = MAX_INT_BYTES // (2 * max_str_len)  # safety factor of 2
        all_chunks = []
        for chunk in array.chunks:
            if len(chunk) < max_elems_per_chunk:
                all_chunks.append(chunk)
            else:
                curr_pos = 0
                total_len = len(chunk)
                while curr_pos < total_len:
                    sliced = chunk.slice(curr_pos, max_elems_per_chunk)
                    curr_pos += len(sliced)
                    all_chunks.append(sliced)
        array = pa.chunked_array(all_chunks, type=dtype)

    return pc.cast(array, pa.string())


def file_to_table(
    path: str,
    content_type: str,
    content_encoding: str = ContentEncoding.IDENTITY.value,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    fs_open_kwargs: Dict[str, Any] = {},
    **kwargs,
) -> pa.Table:
    """
    Read a file into a PyArrow Table using any filesystem.

    Args:
        path: The file path to read
        content_type: The content type of the file (e.g., ContentType.CSV.value)
        content_encoding: The content encoding (default: IDENTITY)
        filesystem: The filesystem to use (if None, will be inferred from path)
        column_names: Optional column names to assign
        include_columns: Optional columns to include in the result
        pa_read_func_kwargs_provider: Optional kwargs provider for customization
        fs_open_kwargs: Optional kwargs for filesystem open operations
        **kwargs: Additional kwargs passed to the reader function

    Returns:
        pa.Table: The loaded PyArrow Table
    """
    logger.debug(
        f"Reading {path} to PyArrow. Content type: {content_type}. "
        f"Encoding: {content_encoding}"
    )

    if (
        content_type == ContentType.PARQUET.value
        and content_encoding == ContentEncoding.IDENTITY.value
        and not filesystem
        and path.startswith("s3://")
    ):
        # Use optimized partial parquet reader for s3 if possible
        logger.debug(
            f"Reading {path} using parquet reader for encoding={content_encoding} "
            f"and content_type={content_type}"
        )

        parquet_reader_func = None
        if kwargs.get(READER_TYPE_KWARG, "daft") == "daft":
            from deltacat.utils.daft import daft_file_to_pyarrow_table

            parquet_reader_func = daft_file_to_pyarrow_table
        elif partial_file_download_params and isinstance(
            partial_file_download_params, PartialParquetParameters
        ):
            parquet_reader_func = partial_parquet_file_to_table

        if parquet_reader_func is not None:
            return parquet_reader_func(
                path=path,
                content_type=content_type,
                content_encoding=content_encoding,
                filesystem=filesystem,
                column_names=column_names,
                include_columns=include_columns,
                pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
                partial_file_download_params=partial_file_download_params,
                **kwargs,
            )

    if READER_TYPE_KWARG in kwargs:
        kwargs.pop(READER_TYPE_KWARG)

    pa_read_func = CONTENT_TYPE_TO_READ_FN.get(content_type)
    if not pa_read_func:
        raise NotImplementedError(
            f"PyArrow reader for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{list(CONTENT_TYPE_TO_READ_FN.keys())}"
        )

    reader_kwargs = content_type_to_reader_kwargs(content_type)

    _add_column_kwargs(content_type, column_names, include_columns, reader_kwargs)

    # Merge with provided kwargs
    reader_kwargs.update(kwargs)

    if pa_read_func_kwargs_provider:
        reader_kwargs = pa_read_func_kwargs_provider(content_type, reader_kwargs)

    logger.debug(f"Reading {path} via {pa_read_func} with kwargs: {reader_kwargs}")

    table, latency = timed_invocation(
        pa_read_func,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        content_encoding=content_encoding,
        **reader_kwargs,
    )
    logger.debug(f"Time to read {path} into PyArrow Table: {latency}s")
    return table


def file_to_parquet(
    path: str,
    content_type: str = ContentType.PARQUET.value,
    content_encoding: str = ContentEncoding.IDENTITY.value,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    fs_open_kwargs: Dict[str, Any] = {},
    **kwargs,
) -> ParquetFile:
    """
    Read a file into a PyArrow ParquetFile using any filesystem.

    It returns a ParquetFile object which provides metadata access and lazy loading.

    Args:
        path: The file path to read
        content_type: The content type (must be PARQUET, default: PARQUET)
        content_encoding: The content encoding (must be IDENTITY, default: IDENTITY)
        filesystem: The filesystem to use (if None, will be inferred from path)
        column_names: Optional column names (unused for ParquetFile but kept for API consistency)
        include_columns: Optional columns (unused for ParquetFile but kept for API consistency)
        pa_read_func_kwargs_provider: Optional kwargs provider for customization
        fs_open_kwargs: Optional kwargs for filesystem open operations
        **kwargs: Additional kwargs passed to ParquetFile constructor

    Returns:
        ParquetFile: The ParquetFile object for lazy loading and metadata access

    Raises:
        ContentTypeValidationError: If content_type is not PARQUET or content_encoding is not IDENTITY
    """
    logger.debug(
        f"Reading {path} to PyArrow ParquetFile. "
        f"Content type: {content_type}. Encoding: {content_encoding}"
    )
    # Validate content type and encoding
    if (
        content_type != ContentType.PARQUET.value
        or content_encoding != ContentEncoding.IDENTITY.value
    ):
        raise ContentTypeValidationError(
            f"File with content type: {content_type} and content encoding: {content_encoding} "
            "cannot be read into pyarrow.parquet.ParquetFile"
        )

    # Resolve filesystem and path
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)

    # Build kwargs for ParquetFile constructor
    parquet_kwargs = {}

    # Add filesystem to kwargs if we have one
    if filesystem:
        parquet_kwargs["filesystem"] = filesystem

    # Apply kwargs provider if provided
    if pa_read_func_kwargs_provider:
        parquet_kwargs = pa_read_func_kwargs_provider(content_type, parquet_kwargs)

    # Merge with provided kwargs
    parquet_kwargs.update(kwargs)

    logger.debug(f"Pre-sanitize kwargs for {path}: {parquet_kwargs}")

    # Sanitize kwargs to only include those supported by ParquetFile.__init__
    parquet_kwargs = sanitize_kwargs_to_callable(ParquetFile.__init__, parquet_kwargs)

    logger.debug(
        f"Reading the file from {path} into ParquetFile with kwargs: {parquet_kwargs}"
    )

    def _create_parquet_file():
        return ParquetFile(path, **parquet_kwargs)

    pq_file, latency = timed_invocation(_create_parquet_file)

    logger.debug(f"Time to get {path} into parquet file: {latency}s")

    return pq_file


def concat_tables(
    tables: List[Union[pa.Table, papq.ParquetFile]],
    promote_options: Optional[str] = "permissive",
    **kwargs,
) -> Optional[Union[pa.Table, List[papq.ParquetFile]]]:
    """
    Concatenate a list of PyArrow Tables or ParquetFiles.

    Args:
        tables: List of PyArrow Tables or ParquetFiles to concatenate

    Returns:
        - Single table/ParquetFile if only one input
        - List of ParquetFiles if all inputs are ParquetFiles (preserves lazy behavior)
        - Concatenated PyArrow Table if mixed types or multiple PyArrow Tables
        - None if input is empty
    """
    if tables is None or not len(tables):
        return None
    if len(tables) == 1:
        # Return single table as-is to preserve lazy behavior
        return next(iter(tables))

    # Check if all tables are ParquetFiles - return list to preserve lazy behavior
    if all(isinstance(table, papq.ParquetFile) for table in tables):
        return list(tables)

    # Convert all tables to PyArrow Tables for concatenation
    converted_tables = []
    for table in tables:
        if isinstance(table, papq.ParquetFile):
            converted_tables.append(table.read())
        else:
            converted_tables.append(table)

    return pa.concat_tables(converted_tables, promote_options=promote_options, **kwargs)


def delta_manifest_to_table(
    manifest: "Manifest",
    delta: Optional["Delta"] = None,
) -> pa.Table:
    """Create a flattened PyArrow table from a delta manifest.

    This implementation can process ~1.4MM records/second on a
    10-core 2025 Macbook Air M4 with 16GB of RAM.

    Args:
        manifest: The manifest to convert to a table
        delta: Optional parent delta of the manifest

    Returns:
        PyArrow table with flattened manifest entry data
    """
    if not manifest.entries:
        return pa.table({})

    num_entries = len(manifest.entries)

    # Get manifest-level data once
    manifest_author = manifest.author
    author_name = manifest_author.name if manifest_author else None
    author_version = manifest_author.version if manifest_author else None

    # Get delta-level data once
    stream_position = delta.stream_position if delta else None
    previous_stream_position = delta.previous_stream_position if delta else None

    # Pre-allocate lists for core columns to avoid repeated list operations
    url_values = [None] * num_entries
    id_values = [None] * num_entries
    mandatory_values = [None] * num_entries

    # Meta columns - most common fields in manifest entries
    meta_record_count = [None] * num_entries
    meta_content_length = [None] * num_entries
    meta_source_content_length = [None] * num_entries
    meta_content_type = [None] * num_entries
    meta_content_encoding = [None] * num_entries

    # Track any additional meta fields we haven't seen before
    additional_meta_fields = {}
    additional_entry_fields = {}

    # Single pass through entries with direct list assignment
    for i, entry in enumerate(manifest.entries):
        # Handle core entry fields efficiently
        url_values[i] = entry.get("url") or entry.get("uri")
        id_values[i] = entry.get("id")
        mandatory_values[i] = entry.get("mandatory")

        # Handle meta fields efficiently
        meta = entry.get("meta", {})
        meta_record_count[i] = meta.get("record_count")
        meta_content_length[i] = meta.get("content_length")
        meta_source_content_length[i] = meta.get("source_content_length")
        meta_content_type[i] = meta.get("content_type")
        meta_content_encoding[i] = meta.get("content_encoding")

        # Handle any additional meta fields not in our core set
        for meta_key, meta_value in meta.items():
            if meta_key not in {
                "record_count",
                "content_length",
                "source_content_length",
                "content_type",
                "content_encoding",
                "entry_type",
            }:
                field_name = f"meta_{meta_key}"
                if field_name not in additional_meta_fields:
                    additional_meta_fields[field_name] = [None] * num_entries
                additional_meta_fields[field_name][i] = meta_value

        # Handle any additional entry fields not in our core set
        for entry_key, entry_value in entry.items():
            if entry_key not in {"url", "uri", "id", "mandatory", "meta"}:
                if entry_key not in additional_entry_fields:
                    additional_entry_fields[entry_key] = [None] * num_entries
                additional_entry_fields[entry_key][i] = entry_value

    # Build the arrays dict with core columns
    arrays_dict = {
        "id": pa.array(id_values),
        "mandatory": pa.array(mandatory_values),
        "meta_content_encoding": pa.array(meta_content_encoding),
        "meta_content_length": pa.array(meta_content_length),
        "meta_content_type": pa.array(meta_content_type),
        "meta_record_count": pa.array(meta_record_count),
        "meta_source_content_length": pa.array(meta_source_content_length),
        "path": pa.array(url_values),
    }

    # Add additional fields if they exist
    for field_name, field_values in additional_meta_fields.items():
        arrays_dict[field_name] = pa.array(field_values)

    for field_name, field_values in additional_entry_fields.items():
        arrays_dict[field_name] = pa.array(field_values)

    # Add manifest/delta columns only if they have data (avoid null columns)
    if author_name is not None:
        arrays_dict["author_name"] = pa.array([author_name] * num_entries)
    if author_version is not None:
        arrays_dict["author_version"] = pa.array([author_version] * num_entries)
    if stream_position is not None:
        arrays_dict["stream_position"] = pa.array([stream_position] * num_entries)
    if previous_stream_position is not None:
        arrays_dict["previous_stream_position"] = pa.array(
            [previous_stream_position] * num_entries
        )

    return pa.table(arrays_dict)


def get_base_arrow_type_name(arrow_type: pa.DataType) -> str:
    """Get the base type name from a PyArrow DataType for compatibility lookup.

    This function normalizes complex PyArrow types to their base type names for
    use in reader compatibility validation. Only specific complex types are
    normalized; all others return their string representation.

    Args:
        arrow_type: The PyArrow DataType to normalize

    Returns:
        str: The normalized type name for compatibility lookup

    Examples:
        >>> get_base_arrow_type_name(pa.int32())
        'int32'
        >>> get_base_arrow_type_name(pa.list_(pa.int32()))
        'list'
        >>> get_base_arrow_type_name(pa.timestamp('s', tz='UTC'))
        'timestamp_tz'
    """
    # Only normalize specific complex types, otherwise return str(arrow_type)
    if isinstance(arrow_type, pa.FixedShapeTensorType):
        return "fixed_shape_tensor"
    elif pa.types.is_large_list(arrow_type):
        return "large_list"
    elif pa.types.is_list_view(arrow_type):
        return "list_view"
    elif pa.types.is_large_list_view(arrow_type):
        return "large_list_view"
    elif pa.types.is_fixed_size_list(arrow_type):
        return "fixed_size_list"
    elif pa.types.is_list(arrow_type):
        return "list"
    elif pa.types.is_map(arrow_type):
        return "map"
    elif pa.types.is_struct(arrow_type):
        return "struct"
    elif pa.types.is_dictionary(arrow_type):
        return "dictionary"
    elif pa.types.is_decimal(arrow_type):
        if isinstance(arrow_type, pa.Decimal128Type):
            return "decimal128"
        elif isinstance(arrow_type, pa.Decimal256Type):
            return "decimal256"
    elif pa.types.is_timestamp(arrow_type):
        # Check if it has timezone info
        if arrow_type.tz is not None:
            return f"timestamp_tz[{arrow_type.unit}]"
        else:
            return str(arrow_type)
    else:
        # For all other types, return the string representation
        return str(arrow_type)


def get_supported_test_types() -> List[Tuple[str, str, List[Any]]]:
    """Get comprehensive PyArrow types supported by DeltaCAT writers and readers.

    This utility function returns example Arrow arrays for every Arrow type
    supported by DeltaCAT writers and readers of tables with schemas. The data
    is used for testing compatibility between different dataset types and
    content types.

    Returns:
        List[Tuple[str, str, List[Any]]]: List of tuples containing:
            - Test name (str): Human-readable name for the test case
            - Arrow type code (str): Python code to create the PyArrow DataType
            - Test data (List[Any]): Sample data values for testing

    Examples:
        >>> test_types = get_supported_test_types()
        >>> for name, type_code, data in test_types[:2]:
        ...     print(f"{name}: {type_code} -> {data}")
        int8: pa.int8() -> [127, -128, 0]
        int16: pa.int16() -> [32767, -32768, 1000]
    """

    return [
        # Integer types
        ("int8", "pa.int8()", [127, -128, 0]),
        ("int16", "pa.int16()", [32767, -32768, 1000]),
        ("int32", "pa.int32()", [2147483647, -2147483648, 1000]),
        ("int64", "pa.int64()", [9223372036854775807, -9223372036854775808, 1000]),
        ("uint8", "pa.uint8()", [255, 0, 128]),
        ("uint16", "pa.uint16()", [65535, 0, 1000]),
        ("uint32", "pa.uint32()", [4294967295, 0, 1000]),
        ("uint64", "pa.uint64()", [18446744073709551615, 0, 1000]),
        # Float types
        ("float16", "pa.float16()", np.array([1.5, np.nan], dtype=np.float16)),
        ("float32", "pa.float32()", [3.14159, -2.71828, 1.41421]),
        ("float64", "pa.float64()", [1.123456789, -2.987654321, 3.141592653589793]),
        # Boolean and null
        ("bool_", "pa.bool_()", [True, False, True]),
        ("null", "pa.null()", [None, None, None]),
        # String types
        ("string", "pa.string()", ["hello", "world", "test"]),
        (
            "large_string",
            "pa.large_string()",
            ["large hello", "large world", "large test"],
        ),
        # Binary types
        ("binary", "pa.binary()", [b"hello", b"world", b"test"]),
        (
            "large_binary",
            "pa.large_binary()",
            [b"large hello", b"large world", b"large test"],
        ),
        # Date and time types
        (
            "date32",
            "pa.date32()",
            [date(2023, 1, 1), date(2023, 12, 31), date(2024, 6, 15)],
        ),
        (
            "date64",
            "pa.date64()",
            [date(2023, 1, 1), date(2023, 12, 31), date(2024, 6, 15)],
        ),
        ("time32_s", "pa.time32('s')", [1754962113, 1754962114, 1754962115]),
        ("time32_ms", "pa.time32('ms')", [1754962113, 1754962114, 1754962115]),
        (
            "time64_us",
            "pa.time64('us')",
            [1754962113000000, 1754962114000000, 1754962115000000],
        ),
        (
            "time64_ns",
            "pa.time64('ns')",
            [1754962113000000000, 1754962114000000000, 1754962115000000000],
        ),
        (
            "timestamp_s",
            "pa.timestamp('s')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_ms",
            "pa.timestamp('ms')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_us",
            "pa.timestamp('us')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_ns",
            "pa.timestamp('ns')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_s_utc",
            "pa.timestamp('s', tz='UTC')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_ms_utc",
            "pa.timestamp('ms', tz='UTC')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_us_utc",
            "pa.timestamp('us', tz='UTC')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        (
            "timestamp_ns_utc",
            "pa.timestamp('ns', tz='UTC')",
            [
                datetime(2023, 1, 1, 12, 0, 0),
                datetime(2023, 12, 31, 23, 59, 59),
                datetime(2024, 6, 15, 10, 30, 45),
            ],
        ),
        ("duration_s", "pa.duration('s')", [1754962113, 1754962114, 1754962115]),
        (
            "duration_ms",
            "pa.duration('ms')",
            [1754962113000, 1754962114000, 1754962115000],
        ),
        (
            "duration_us",
            "pa.duration('us')",
            [1754962113000000, 1754962114000000, 1754962115000000],
        ),
        (
            "duration_ns",
            "pa.duration('ns')",
            [1754962113000000000, 1754962114000000000, 1754962115000000000],
        ),
        (
            "month_day_nano",
            "pa.month_day_nano_interval()",
            [
                pa.scalar((1, 15, -30), type=pa.month_day_nano_interval()),
                pa.scalar((2, 15, -30), type=pa.month_day_nano_interval()),
                pa.scalar((3, 15, -30), type=pa.month_day_nano_interval()),
            ],
        ),
        # Decimal
        (
            "decimal128_5_2",
            "pa.decimal128(5, 2)",
            [Decimal("123.45"), Decimal("-67.89"), Decimal("999.99")],
        ),
        (
            "decimal128_38_0",
            "pa.decimal128(38, 0)",
            [
                Decimal("12345678901234567890123456789012345678"),
                Decimal("-12345678901234567890123456789012345678"),
                Decimal("0"),
            ],
        ),
        (
            "decimal128_1_0",
            "pa.decimal128(1, 0)",
            [Decimal("1"), Decimal("2"), Decimal("3")],
        ),
        (
            "decimal128_38_10",
            "pa.decimal128(38, 10)",
            [
                Decimal("1234567890123456789012345678.9012345678"),
                Decimal("-1234567890123456789012345678.9012345678"),
                Decimal("0.0000000000"),
            ],
        ),
        (
            "decimal256_76_0",
            "pa.decimal256(76, 0)",
            [
                Decimal(
                    "1234567890123456789012345678901234567812345678901234567890123456789012345678"
                ),
                Decimal("-0"),
                Decimal("0"),
            ],
        ),
        (
            "decimal256_1_0",
            "pa.decimal256(1, 0)",
            [Decimal("1"), Decimal("2"), Decimal("3")],
        ),
        (
            "decimal256_5_2",
            "pa.decimal256(5, 2)",
            [Decimal("123.45"), Decimal("-67.89"), Decimal("999.99")],
        ),
        (
            "decimal256_76_38",
            "pa.decimal256(76, 38)",
            [
                Decimal(
                    "12345678901234567890123456789012345678.12345678901234567890123456789012345678"
                ),
                Decimal("-0.00000000000000000000000000000000000000"),
                Decimal("0.00000000000000000000000000000000000000"),
            ],
        ),
        # List types
        ("list_int32", "pa.list_(pa.int32())", [[1, 2, 3], [4, 5], [6, 7, 8, 9]]),
        ("list_string", "pa.list_(pa.string())", [["a", "b"], ["c", "d", "e"], ["f"]]),
        # Struct type
        (
            "struct_simple",
            "pa.struct([('name', pa.string()), ('age', pa.int32())])",
            [
                {"name": "Alice", "age": 30},
                {"name": "Bob", "age": 25},
                {"name": "Charlie", "age": 35},
            ],
        ),
        (
            "large_list_int32",
            "pa.large_list(pa.int32())",
            [[1, 2, 3], [4, 5], [6, 7, 8, 9]],
        ),
        (
            "fixed_size_list_int32",
            "pa.list_(pa.int32(), 3)",
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        ),
        (
            "list_view_int32",
            "pa.list_view(pa.int32())",
            [[1, 2, 3], [4, 5], [6, 7, 8, 9]],
        ),
        (
            "large_list_view_int32",
            "pa.large_list_view(pa.int32())",
            [[1, 2, 3], [4, 5], [6, 7, 8, 9]],
        ),
        # Dictionary type
        (
            "dictionary_string",
            "pa.dictionary(pa.int32(), pa.string())",
            ["apple", "banana", "apple"],
        ),
        # Map type
        (
            "map_string_int32",
            "pa.map_(pa.string(), pa.int32())",
            [{"a": 1, "b": 2}, {"c": 3, "d": 4}, {"e": 5}],
        ),
        # Extension Types
        (
            "fixed_shape_tensor",
            "pa.fixed_shape_tensor(pa.int32(), [3, 3])",
            [
                np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.int32),
                np.array([10, 11, 12, 13, 14, 15, 16, 17, 18], dtype=np.int32),
                np.array([19, 20, 21, 22, 23, 24, 25, 26, 27], dtype=np.int32),
            ],
        ),
    ]



import importlib
from contextlib import nullcontext
import logging
import time
import ray

import deltacat
from deltacat.compute.compactor import (
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat import logs
from deltacat.compute.compactor_v2.model.evaluate_compaction_result import (
    ExecutionCompactionResult,
)
from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.compactor_v2.deletes.delete_strategy import (
    DeleteStrategy,
)
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.compactor_v2.model.merge_result import MergeResult
from deltacat.compute.compactor_v2.deletes.delete_file_envelope import (
    DeleteFileEnvelope,
)
from deltacat.storage import (
    Delta,
    DeltaLocator,
    PartitionLocator,
)
from deltacat.storage.model.manifest import Manifest
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.resources import (
    get_current_process_peak_memory_usage_in_bytes,
)
from deltacat.compute.compactor_v2.private.compaction_utils import (
    _get_rci_source_partition_locator,
    _fetch_compaction_metadata,
    _build_uniform_deltas,
    _group_uniform_deltas,
    _stage_new_partition,
    _run_hash_and_merge,
    _process_merge_results,
    _create_round_completion_info,
    _commit_compaction_result,
)
from deltacat.utils.metrics import metrics
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)

from typing import List, Optional
from deltacat.compute.compactor_v2.utils import io
from deltacat.exceptions import categorize_errors
from deltacat.compute.compactor_v2.constants import COMPACT_PARTITION_METRIC_PREFIX

if importlib.util.find_spec("memray"):
    import memray


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@metrics(prefix=COMPACT_PARTITION_METRIC_PREFIX)
@categorize_errors
def compact_partition(params: CompactPartitionParams, **kwargs) -> None:
    assert (
        params.hash_bucket_count is not None and params.hash_bucket_count >= 1
    ), "hash_bucket_count is a required arg for compactor v2"
    assert type(params.hash_bucket_count) is int, "Hash bucket count must be an integer"
    if params.num_rounds > 1:
        assert (
            not params.drop_duplicates
        ), "num_rounds > 1, drop_duplicates must be False but is True"

    with (
        memray.Tracker("compaction_partition.bin")
        if params.enable_profiler
        else nullcontext()
    ):
        execute_compaction_result: ExecutionCompactionResult = _execute_compaction(
            params,
            **kwargs,
        )
        _commit_compaction_result(params, execute_compaction_result)


def _execute_compaction(
    params: CompactPartitionParams, **kwargs
) -> ExecutionCompactionResult:
    compaction_start_time: float = time.monotonic()
    # Fetch round completion info for previously compacted partition, if it exists
    fetch_compaction_metadata_result: tuple[
        Optional[Manifest], Optional[RoundCompletionInfo]
    ] = _fetch_compaction_metadata(params)
    (
        previous_compacted_delta_manifest,
        round_completion_info,
    ) = fetch_compaction_metadata_result
    rci_source_partition_locator: PartitionLocator = _get_rci_source_partition_locator(
        params
    )

    base_audit_url: str = rci_source_partition_locator.path(
        f"{params.compaction_artifact_path}/compaction-audit"
    )
    audit_url: str = f"{base_audit_url}.json"
    logger.info(f"Compaction audit will be written to {audit_url}")
    compaction_audit: CompactionSessionAuditInfo = (
        CompactionSessionAuditInfo(deltacat.__version__, ray.__version__, audit_url)
        .set_hash_bucket_count(params.hash_bucket_count)
        .set_compactor_version(CompactorVersion.V2.value)
    )

    if params.pg_config:
        logger.info(
            "pg_config specified. Tasks will be scheduled in a placement group."
        )
        cluster_resources = params.pg_config.resource
        cluster_memory = cluster_resources["memory"]
        compaction_audit.set_total_cluster_memory_bytes(cluster_memory)
    high_watermark = (
        round_completion_info.high_watermark if round_completion_info else None
    )
    audit_url = compaction_audit.audit_url if compaction_audit else None
    # discover and build uniform deltas
    delta_discovery_start = time.monotonic()
    input_deltas: List[Delta] = io.discover_deltas(
        params.source_partition_locator,
        params.last_stream_position_to_compact,
        params.rebase_source_partition_locator,
        params.rebase_source_partition_high_watermark,
        high_watermark,
        params.deltacat_storage,
        params.deltacat_storage_kwargs,
        params.list_deltas_kwargs,
    )
    if not input_deltas:
        logger.info("No input deltas found to compact.")
        return ExecutionCompactionResult(None, None, False)
    build_uniform_deltas_result: tuple[
        List[DeltaAnnotated], DeleteStrategy, List[DeleteFileEnvelope]
    ] = _build_uniform_deltas(
        params, compaction_audit, input_deltas, delta_discovery_start
    )
    (
        uniform_deltas,
        delete_strategy,
        delete_file_envelopes,
    ) = build_uniform_deltas_result
    logger.info(f"Number of rounds parameter is set to: {params.num_rounds}")
    uniform_deltas_grouped = _group_uniform_deltas(params, uniform_deltas)
    logger.info(f"Length of grouped uniform deltas is: {len(uniform_deltas_grouped)}")
    merge_result_list: List[MergeResult] = []
    compacted_partition = _stage_new_partition(params)
    for uniform_deltas in uniform_deltas_grouped:
        # run hash and merge
        _run_hash_and_merge_result: List[MergeResult] = _run_hash_and_merge(
            params,
            uniform_deltas,
            round_completion_info,
            delete_strategy,
            delete_file_envelopes,
            compaction_audit,
            previous_compacted_delta_manifest,
            compacted_partition,
        )
        merge_result_list.extend(_run_hash_and_merge_result)
    # process merge results
    process_merge_results: tuple[
        Delta, list[MaterializeResult], dict
    ] = _process_merge_results(params, merge_result_list, compaction_audit)
    merged_delta, mat_results, hb_id_to_entry_indices_range = process_merge_results
    # Record information, logging, and return ExecutionCompactionResult
    record_info_msg: str = f" Materialized records: {merged_delta.meta.record_count}"
    logger.info(record_info_msg)
    compacted_delta: Delta = params.deltacat_storage.commit_delta(
        merged_delta,
        properties=kwargs.get("properties", {}),
        **params.deltacat_storage_kwargs,
    )

    logger.info(f"Committed compacted delta: {compacted_delta}")
    compaction_end_time: float = time.monotonic()
    compaction_audit.set_compaction_time_in_seconds(
        compaction_end_time - compaction_start_time
    )
    new_compacted_delta_locator: DeltaLocator = DeltaLocator.of(
        compacted_partition.locator,
        compacted_delta.stream_position,
    )
    pyarrow_write_result: PyArrowWriteResult = PyArrowWriteResult.union(
        [m.pyarrow_write_result for m in mat_results]
    )

    session_peak_memory = get_current_process_peak_memory_usage_in_bytes()
    compaction_audit.set_peak_memory_used_bytes_by_compaction_session_process(
        session_peak_memory
    )

    compaction_audit.save_round_completion_stats(mat_results)

    compaction_result: ExecutionCompactionResult = _create_round_completion_info(
        params,
        compaction_audit,
        compacted_partition,
        audit_url,
        hb_id_to_entry_indices_range,
        rci_source_partition_locator,
        new_compacted_delta_locator,
        pyarrow_write_result,
        round_completion_info,
    )
    return compaction_result



from __future__ import annotations

from deltacat.experimental.storage.rivulet.parquet.serializer import (
    ParquetDataSerializer,
)
from deltacat.experimental.storage.rivulet import Schema
from deltacat.experimental.storage.rivulet.serializer import DataSerializer
from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider

from deltacat.experimental.storage.rivulet.feather.serializer import (
    FeatherDataSerializer,
)


class DataSerializerFactory:
    """
    Simple factory class for getting the appropriate serializer given a schema
    TODO make this more modular/pluggable like DatasetReaderRegistrar
      This will be more challenging to make pluggable, because we should not rely on a simple 1:1 mapping of type to serializer
      The actual logic for determining how to serialize a given schema may be complex
      e.g.: if schema contains datatype X, you must use serializer Y. Otherwise, default to serializer Z
    """

    @classmethod
    def get_serializer(
        self,
        schema: Schema,
        file_provider: FileProvider,
        user_provided_format: str | None = None,
    ) -> DataSerializer:
        if user_provided_format == "parquet":
            return ParquetDataSerializer(file_provider, schema)
        elif user_provided_format == "feather":
            return FeatherDataSerializer(file_provider, schema)
        elif user_provided_format is not None:
            raise ValueError("Unsupported format. Must be 'parquet' or 'feather'.")

        # Default engine logic. For now, if there is image or binary use feather
        has_binary_or_image = any(
            field.datatype.type_name.startswith(("binary", "image"))
            for field in schema.values()
        )
        if has_binary_or_image:
            return FeatherDataSerializer(file_provider, schema)
        else:
            return ParquetDataSerializer(file_provider, schema)



from __future__ import annotations

from typing import List, Callable, Any

from deltacat.experimental.storage.rivulet.mvp.Table import MvpTable
from deltacat.experimental.storage.rivulet import Schema
from deltacat.experimental.storage.rivulet.reader.data_scan import DataScan
from deltacat.experimental.storage.rivulet.reader.dataset_metastore import (
    DatasetMetastore,
)
from deltacat.experimental.storage.rivulet.reader.dataset_reader import DatasetReader
from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)


class DatasetExecutor:
    """
    Executor class which runs operations such as select, map, take, save

    This class may store intermediate state while it is executing.

    LogicalPlan is responsible for constructor an executor and ordering operations appropriately
    """

    def __init__(
        self,
        schema: Schema,
        metastore: DatasetMetastore,
    ):
        self.effective_schema: Schema = schema.__deepcopy__()
        self.output: MvpTable | None = None
        self._metastore = metastore

    def collect(self) -> MvpTable:
        if not self.output:
            self.output = self._read(self.effective_schema)
        return self.output

    def select(self, fields: List[str]) -> "DatasetExecutor":
        """
        Reads data and selects a subset of fields
        Note that this implementation is super inefficient (does not push down filters to read, copies data to new MvpTable). That is OK since this will all be replaced
        """
        # Read data from original input sources if not already read
        if not self.output:
            self.output = self._read(self.effective_schema)
        # Calculate effective schema and apply it to data
        self.effective_schema.filter(fields)
        self.output = MvpTable(
            {
                key: value
                for key, value in self.output.data.items()
                if key in self.effective_schema
            },
        )
        return self

    def map(self, transform: Callable[[Any], Any]) -> "DatasetExecutor":
        raise NotImplementedError

    def _read(self, schema: Schema) -> MvpTable:
        """
        Internal helper method to read data

        TODO for now this is doing dumb in-memory implementation and later this is going to be replaced by rust library
        """
        return self._read_as_mvp_table(schema)

    def _read_as_mvp_table(self, schema: Schema):
        data = list(
            DataScan(
                schema, QueryExpression(), DatasetReader(self._metastore)
            ).to_pydict()
        )
        output = {}
        for key in schema.fields.keys():
            output[key] = [d.get(key) for d in data]
        return MvpTable(output)



from __future__ import annotations

from typing import Tuple, Optional

from deltacat.catalog.model.catalog import Catalog as DCCatalog
from deltacat.catalog.model.table_definition import TableDefinition

from daft.catalog import Catalog, Identifier, Table
from daft.dataframe import DataFrame
from daft.logical.schema import Schema
from deltacat.constants import DEFAULT_NAMESPACE


class DaftCatalog(Catalog):
    """
    Wrapper class to create a Daft catalog from a DeltaCAT catalog.

    The initialization of DeltaCAT and Daft catalogs is managed in `deltacat.catalog.catalog.py`. The user
    is just expected to initialize catalogs through the DeltaCAT public interface (init / put_catalog).

    TODO (mccember) in follow up PR we need to consider how to keep the DeltaCAT Catalogs class and Daft session in sync,
      and the user-facing entrypoint to get a Daft catalog

    This class itself expects a `Catalog` and will invoke the underlying implementation
    similar to `deltacat.catalog.delegate.py`, like:
      catalog.impl.create_namespace(namespace, inner=catalog.inner)

    We cannot route calls through the higher level catalog registry / delegate since this wrapper class is at a lower
     layer and does not manage registering catalogs.
    """

    def __init__(self, catalog: DCCatalog, name: str):
        """
        Initialize given DeltaCAT catalog. This catalog is also registered with DeltaCAT (via deltacat.put_catalog) given the provided Name

        :param catalog: DeltaCAT Catalog object. If None, the catalog will be fetched from `deltacat.Catalogs`
            given the catalog name.

        :param name: Name of DeltaCAT catalog. If the name is not yet registered with `deltacat.Catalogs`,
            it will be registered upon creation to ensure that the DeltaCAT and Daft catalogs keep in sync.

        :param kwargs: Additional keyword arguments passed to deltacat.get_catalog or deltacat.put_catalog,
                       such as 'namespace' for tests.
        """
        self.dc_catalog = catalog
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    ###
    # create_*
    ###
    def create_namespace(self, identifier: Identifier | str):
        """Create a new namespace in the catalog."""
        if isinstance(identifier, Identifier):
            identifier = str(identifier)
        self.dc_catalog.impl.create_namespace(identifier, inner=self.dc_catalog.inner)

    def create_table(
        self, identifier: Identifier | str, source: Schema | DataFrame, **kwargs
    ) -> Table:
        """
        Create a DeltaCAT table via Daft catalog API

        End users calling create_table through the daft table API may provide kwargs which will be plumbed through
        to deltacat create_table. For full list of keyword arguments accepted by create_table.

        Note: as of 4/22, Daft create_table does not yet support kwargs. Tracked at: https://github.com/Eventual-Inc/Daft/issues/4195

        :param identifier: Daft table identifier. Sequence of strings of the format (namespace) or (namespace, table)
            or (namespace, table, table version). If this is a string, it is a dot delimited string of the same format.
            Identifiers can be created either like Identifier("namespace", "table", "version") OR
                Identifier.from_str("namespace.table.version")

        :param source: a TableSource, either a Daft DataFrame, Daft Schema, or str (filesystem path)
        """
        if isinstance(source, DataFrame):
            return self._create_table_from_df(identifier, source)
        elif isinstance(source, Schema):
            return self._create_table_from_schema(identifier, source)
        else:
            raise Exception(
                f"Expected table source to be Schema or DataFrame. Found: {type(source)}"
            )

    def _create_table_from_df(
        self, ident: Identifier | str, source: DataFrame, **kwargs
    ) -> Table:
        """
        Create a table from a DataFrame.
        """
        t = self._create_table_from_schema(ident, source.schema(), **kwargs)
        # TODO (mccember) append data upon creation
        return t

    def _create_table_from_schema(
        self, ident: Identifier | str, source: Schema, **kwargs
    ) -> Table:
        """
        Create a table from a schema.
        """
        namespace, name, version = self._extract_namespace_name_version(ident)

        # Convert the Daft schema to a DeltaCAT schema
        # This is a simplified version, would need to be enhanced for production
        deltacat_schema = self._convert_schema_to_deltacat(source)

        # Create the table in DeltaCAT
        table_def = self.dc_catalog.impl.create_table(
            name,
            namespace=namespace,
            version=version,
            schema=deltacat_schema,
            inner=self.dc_catalog.inner,
            **kwargs,
        )

        return DaftTable._from_obj(table_def)

    ###
    # drop_*
    ###

    def drop_namespace(self, identifier: Identifier | str):
        raise NotImplementedError()

    def drop_table(self, identifier: Identifier | str):
        raise NotImplementedError()

    ###
    # get_*
    ###

    def get_table(self, identifier: Identifier | str, **kwargs) -> Table:
        namespace, table, version = self._extract_namespace_name_version(identifier)

        table_def = self.dc_catalog.impl.get_table(
            table,
            namespace=namespace,
            table_version=version,
            inner=self.dc_catalog.inner,
            **kwargs,
        )

        if not table_def:
            raise ValueError(f"Table {identifier} not found")

        return DaftTable._from_obj(table_def)

    ###
    # list_*
    ###

    def list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        raise NotImplementedError("Not implemented")

    def list_tables(self, pattern: str | None = None) -> list[str]:
        raise NotImplementedError("Not implemented")

    def _extract_namespace_name_version(
        self, ident: Identifier | str
    ) -> Tuple[str, str, Optional[str]]:
        """
        Extract namespace, name,version from identifier

        Returns a 3-tuple. If no namespace is provided, uses DeltaCAT defualt namespace
        """
        default_namespace = DEFAULT_NAMESPACE

        if isinstance(ident, str):
            ident = Identifier.from_str(ident)

        if isinstance(ident, Identifier):
            if len(ident) == 1:
                return (default_namespace, ident[0], None)
            elif len(ident) == 2:
                return (ident[0], ident[1], None)
            elif len(ident) == 3:
                return (ident[0], ident[1], ident[2])
            else:
                raise ValueError(
                    f"Expected table identifier to be in format (table) or (namespace, table)"
                    f"or (namespace, table, version). Found: {ident}"
                )

    def _convert_schema_to_deltacat(self, schema: Schema):
        """Convert Daft schema to DeltaCAT schema.
        For now, just use PyArrow schema as intermediary
        TODO look into how enhancements on schema can be propagated between Daft<=>DeltaCAT
        """
        from deltacat.storage.model.schema import Schema as DeltaCATSchema

        return DeltaCATSchema.of(schema=schema.to_pyarrow_schema())


class DaftTable(Table):
    """
    Wrapper class to create a Daft table from a DeltaCAT table
    """

    _inner: TableDefinition

    _read_options = set()
    _write_options = set()

    def __init__(self, inner: TableDefinition):
        self._inner = inner

    @property
    def name(self) -> str:
        """Return the table name."""
        return self._inner.table_version.table_name

    @staticmethod
    def _from_obj(obj: object) -> DaftTable:
        """Returns a DeltaCATTable if the given object can be adapted so."""
        if isinstance(obj, TableDefinition):
            t = DaftTable.__new__(DaftTable)
            t._inner = obj
            return t
        raise ValueError(f"Unsupported DeltaCAT table type: {type(obj)}")

    def read(self, **options) -> DataFrame:
        raise NotImplementedError("Not implemented")

    def write(self, df: DataFrame | object, mode: str = "append", **options):
        raise NotImplementedError("Not implemented")



from __future__ import annotations

import logging
import itertools

from enum import Enum
from typing import Optional, List, Dict, Any, TYPE_CHECKING
from uuid import uuid4

if TYPE_CHECKING:
    from deltacat.storage.model.schema import FieldLocator

from deltacat import logs

from deltacat.types.media import (
    ContentType,
    ContentEncoding,
    EXT_TO_CONTENT_TYPE,
    EXT_TO_CONTENT_ENCODING,
)

import json
import pyarrow as pa
import posixpath

from deltacat.utils.filesystem import get_file_info

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class EntryType(str, Enum):
    """
    Enum representing all possible content categories of a manifest entry file.

    DATA: The entry contains fully qualified records compliant with the parent
    table's schema to insert and/or update. Data files for upsert Deltas use
    this entry's parameters to find matching fields to update. If no entry
    parameters are specified, then the parent table's primary keys are used.
    Only records from entries in Deltas with lower stream positions than this
    entry will be targeted for update.

    POSITIONAL_DELETE: The entry contains pointers to records in other entries
    to delete. Deleted records will be filtered from query results at runtime.

    EQUALITY_DELETE: The entry contains a subset of field values from the
    table records to find and delete. The full record of any matching data
    entries in Deltas with a lower stream position than this entry's Delta
    will be deleted. The fields used for record discovery are controlled by
    this entry's parameters. If no entry parameters are specified, then the
    fields used for record discovery are linked to the parent table's merge
    keys. The entry may contain additional fields not used for delete record
    discovery which will be ignored. Deleted records will be filtered from
    query results at runtime.
    """

    DATA = "data"
    POSITIONAL_DELETE = "positional_delete"
    EQUALITY_DELETE = "equality_delete"

    @classmethod
    def get_default(cls):
        return EntryType.DATA

    @classmethod
    def list(cls):
        return [c.value for c in EntryType]


class EntryParams(dict):
    """
    Parameters that control manifest entry interpretation.

    For EQUALITY_DELETE manifest entry types, parameters include equality
    field identifiers.
    """

    @staticmethod
    def of(
        equality_field_locators: Optional[List["FieldLocator"]] = None,
    ) -> EntryParams:
        params = EntryParams()
        if equality_field_locators is not None:
            params["equality_field_locators"] = equality_field_locators
        return params

    @property
    def equality_field_locators(self) -> Optional[List["FieldLocator"]]:
        return self.get("equality_field_locators")


class Manifest(dict):
    """
    A DeltaCAT manifest contains metadata common to multiple manifest formats
    like Amazon Redshift and Apache Iceberg to simplify dataset import/export.
    """

    @staticmethod
    def _build_manifest(
        meta: Optional[ManifestMeta],
        entries: Optional[ManifestEntryList],
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        manifest = Manifest()
        manifest["id"] = uuid
        if meta is not None:
            manifest["meta"] = meta
        if entries is not None:
            manifest["entries"] = entries
        if author is not None:
            manifest["author"] = author
        return manifest

    @staticmethod
    def of(
        entries: ManifestEntryList,
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
        entry_type: Optional[EntryType] = None,
        entry_params: Optional[EntryParams] = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        total_record_count = 0
        total_content_length = 0
        total_source_content_length = 0
        content_type = None
        content_encoding = None
        credentials = None
        content_type_params = None
        schema_id = None
        sort_scheme_id = None
        if entries:
            content_type = entries[0].meta.content_type
            content_encoding = entries[0].meta.content_encoding
            credentials = entries[0].meta.credentials
            content_type_params = entries[0].meta.content_type_parameters

            # Keep the latest schema ID
            # Schema IDs are >= 0, and schema evolution always increments the last schema ID
            entry_schema_ids = [
                entry.meta.schema_id if entry.meta.schema_id is not None else -1
                for entry in entries
            ]
            max_schema_id = max(entry_schema_ids) if entry_schema_ids else -1
            schema_id = max_schema_id if max_schema_id >= 0 else None

            # Handle sort_scheme_id: set to None if entries have multiple different sort_scheme_ids
            entry_sort_scheme_ids = set(
                entry.meta.sort_scheme_id
                for entry in entries
                if entry.meta.sort_scheme_id is not None
            )
            sort_scheme_id = (
                list(entry_sort_scheme_ids)[0]
                if len(entry_sort_scheme_ids) == 1
                else None
            )

            for entry in entries:
                meta = entry.meta
                if meta.content_type != content_type:
                    content_type = None
                if meta.content_encoding != content_encoding:
                    content_encoding = None
                entry_content_type = meta.content_type
                if content_type and entry_content_type != content_type:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"type '{content_type}' but found "
                        f"'{entry_content_type}'"
                    )
                    raise ValueError(msg)
                entry_content_encoding = meta.get("content_encoding", None)
                if content_encoding and entry_content_encoding != content_encoding:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"encoding '{content_encoding}' but found "
                        f"'{entry_content_encoding}'"
                    )
                    raise ValueError(msg)
                actual_entry_type = meta.entry_type
                if entry_type and (actual_entry_type != entry_type):
                    msg = (
                        f"Expected all manifest entries to have type "
                        f"'{entry_type}' but found '{actual_entry_type}'"
                    )
                    raise ValueError(msg)
                actual_entry_params = meta.entry_params
                if entry_params and (actual_entry_params != entry_params):
                    msg = (
                        f"Expected all manifest entries to have params "
                        f"'{entry_params}' but found '{actual_entry_params}'"
                    )
                    raise ValueError(msg)
                actual_credentials = meta.credentials
                if credentials and (actual_credentials != credentials):
                    msg = (
                        f"Expected all manifest entries to have credentials "
                        f"'{credentials}' but found '{actual_credentials}'"
                    )
                    raise ValueError(msg)
                actual_content_type_params = meta.content_type_parameters
                if content_type_params and (
                    actual_content_type_params != content_type_params
                ):
                    msg = (
                        f"Expected all manifest entries to have content type params "
                        f"'{content_type_params}' but found '{actual_content_type_params}'"
                    )
                    raise ValueError(msg)

                total_record_count += meta.record_count or 0
                total_content_length += meta.content_length or 0
                total_source_content_length += meta.source_content_length or 0

        meta = ManifestMeta.of(
            record_count=total_record_count,
            content_length=total_content_length,
            content_type=content_type,
            content_encoding=content_encoding,
            source_content_length=total_source_content_length,
            credentials=credentials,
            content_type_parameters=content_type_params,
            entry_type=entry_type,
            entry_params=entry_params,
            schema_id=schema_id,
            sort_scheme_id=sort_scheme_id,
        )
        manifest = Manifest._build_manifest(meta, entries, author, uuid)
        return manifest

    @staticmethod
    def from_json(json_string: str) -> Manifest:
        parsed_dict = json.loads(json_string)
        return Manifest.of(
            entries=ManifestEntryList.of(
                [
                    ManifestEntry.from_dict(entry)
                    for entry in parsed_dict.get("entries", [])
                ]
            ),
            author=ManifestAuthor.from_dict(parsed_dict.get("author")),
            uuid=parsed_dict.get("id"),
        )

    @staticmethod
    def merge_manifests(
        manifests: List[Manifest], author: Optional[ManifestAuthor] = None
    ) -> Manifest:
        all_entries = ManifestEntryList(
            itertools.chain(*[m.entries for m in manifests])
        )
        merged_manifest = Manifest.of(all_entries, author)
        return merged_manifest

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def entries(self) -> Optional[ManifestEntryList]:
        val: List[ManifestEntry] = self.get("entries")
        if val is not None and not isinstance(val, ManifestEntryList):
            self["entries"] = val = ManifestEntryList.of(val)
        return val

    @property
    def id(self) -> str:
        return self["id"]

    @property
    def author(self) -> Optional[ManifestAuthor]:
        val: Dict[str, Any] = self.get("author")
        if val is not None and not isinstance(val, ManifestAuthor):
            self["author"] = val = ManifestAuthor(val)
        return val


class ManifestMeta(dict):
    @staticmethod
    def of(
        record_count: Optional[int],
        content_length: Optional[int],
        content_type: Optional[str],
        content_encoding: Optional[str],
        source_content_length: Optional[int] = None,
        credentials: Optional[Dict[str, str]] = None,
        content_type_parameters: Optional[List[Dict[str, str]]] = None,
        entry_type: Optional[EntryType] = None,
        entry_params: Optional[EntryParams] = None,
        schema_id: Optional[int] = None,
        sort_scheme_id: Optional[str] = None,
    ) -> ManifestMeta:
        manifest_meta = ManifestMeta()
        if record_count is not None:
            manifest_meta["record_count"] = record_count
        if content_length is not None:
            manifest_meta["content_length"] = content_length
        if source_content_length is not None:
            manifest_meta["source_content_length"] = source_content_length
        if content_type is not None:
            manifest_meta["content_type"] = content_type
        if content_type_parameters is not None:
            manifest_meta["content_type_parameters"] = content_type_parameters
        if content_encoding is not None:
            manifest_meta["content_encoding"] = content_encoding
        if credentials is not None:
            manifest_meta["credentials"] = credentials
        if entry_type is not None:
            manifest_meta["entry_type"] = (
                entry_type.value if isinstance(entry_type, EntryType) else entry_type
            )
        if entry_params is not None:
            manifest_meta["entry_params"] = entry_params
        if schema_id is not None:
            manifest_meta["schema_id"] = schema_id
        if sort_scheme_id is not None:
            manifest_meta["sort_scheme_id"] = sort_scheme_id
        return manifest_meta

    @staticmethod
    def from_dict(obj: dict) -> Optional[ManifestMeta]:
        if obj is None:
            return None

        return ManifestMeta.of(
            record_count=obj.get("record_count"),
            content_length=obj.get("content_length"),
            content_type=obj.get("content_type"),
            content_encoding=obj.get("content_encoding"),
            source_content_length=obj.get("source_content_length"),
            credentials=obj.get("credentials"),
            content_type_parameters=obj.get("content_type_parameters"),
            entry_type=obj.get("entry_type"),
            entry_params=obj.get("entry_params"),
            schema_id=obj.get("schema_id"),
            sort_scheme_id=obj.get("sort_scheme_id"),
        )

    @property
    def record_count(self) -> Optional[int]:
        return self.get("record_count")

    @property
    def content_length(self) -> Optional[int]:
        return self.get("content_length")

    @property
    def content_type(self) -> Optional[str]:
        return self.get("content_type")

    @property
    def content_encoding(self) -> Optional[str]:
        return self.get("content_encoding")

    @property
    def source_content_length(self) -> Optional[int]:
        return self.get("source_content_length")

    @property
    def content_type_parameters(self) -> Optional[List[Dict[str, str]]]:
        return self.get("content_type_parameters")

    @content_type_parameters.setter
    def content_type_parameters(self, params: List[Dict[str, str]]) -> None:
        self["content_type_parameters"] = params

    @property
    def credentials(self) -> Optional[Dict[str, str]]:
        return self.get("credentials")

    @property
    def entry_type(self) -> Optional[EntryType]:
        val = self.get("entry_type")
        if val is not None:
            return EntryType(self["entry_type"])
        return val

    @property
    def entry_params(self) -> Optional[EntryParams]:
        val: Dict[str, Any] = self.get("entry_params")
        if val is not None and not isinstance(val, EntryParams):
            self["entry_params"] = val = EntryParams(val)
        return val

    @property
    def schema_id(self) -> Optional[int]:
        return self.get("schema_id")

    @property
    def sort_scheme_id(self) -> Optional[str]:
        return self.get("sort_scheme_id")


class ManifestEntry(dict):
    @staticmethod
    def of(
        url: Optional[str],
        meta: Optional[ManifestMeta],
        mandatory: bool = True,
        uri: Optional[str] = None,
        uuid: Optional[str] = None,
    ) -> ManifestEntry:
        manifest_entry = ManifestEntry()
        if not (uri or url):
            raise ValueError("No URI or URL specified for manifest entry contents.")
        if (uri and url) and (uri != url):
            raise ValueError(f"Manifest entry URI ({uri}) != URL ({url})")
        if url:
            manifest_entry["url"] = manifest_entry["uri"] = url
        elif uri:
            manifest_entry["url"] = manifest_entry["uri"] = uri
        if meta is not None:
            manifest_entry["meta"] = meta
        if mandatory is not None:
            manifest_entry["mandatory"] = mandatory
        if uuid is not None:
            manifest_entry["id"] = uuid
        return manifest_entry

    @staticmethod
    def from_s3_obj_url(
        url: str,
        record_count: int,
        source_content_length: Optional[int] = None,
        credentials: Optional[Dict[str, str]] = None,
        content_type_parameters: Optional[List[Dict[str, str]]] = None,
        entry_type: Optional[EntryType] = None,
        entry_params: Optional[EntryParams] = None,
        **s3_client_kwargs,
    ) -> ManifestEntry:
        from deltacat.aws import s3u as s3_utils

        s3_obj = s3_utils.get_object_at_url(url, **s3_client_kwargs)
        logger.debug(f"Building manifest entry from {url}: {s3_obj}")
        manifest_entry_meta = ManifestMeta.of(
            record_count=record_count,
            content_length=s3_obj["ContentLength"],
            content_type=s3_obj["ContentType"],
            content_encoding=s3_obj["ContentEncoding"],
            source_content_length=source_content_length,
            credentials=credentials,
            content_type_parameters=content_type_parameters,
            entry_type=entry_type,
            entry_params=entry_params,
        )
        manifest_entry = ManifestEntry.of(url, manifest_entry_meta)
        return manifest_entry

    @staticmethod
    def from_dict(obj: dict) -> ManifestEntry:
        return ManifestEntry.of(
            url=obj.get("url"),
            uri=obj.get("uri"),
            meta=ManifestMeta.from_dict(obj.get("meta")),
            mandatory=obj.get("mandatory", True),
            uuid=obj.get("id"),
        )

    @staticmethod
    def from_path(
        path: str,
        filesystem: pa.fs.FileSystem,
        record_count: int,
        source_content_length: Optional[int] = None,
        content_type: Optional[str] = None,
        content_encoding: Optional[str] = None,
        credentials: Optional[Dict[str, str]] = None,
        content_type_parameters: Optional[List[Dict[str, str]]] = None,
        entry_type: Optional[EntryType] = None,
        entry_params: Optional[EntryParams] = None,
        schema_id: Optional[int] = None,
        sort_scheme_id: Optional[str] = None,
    ) -> ManifestEntry:
        """
        Creates a manifest entry from a path using a pyarrow filesystem.

        Args:
            path: Path to the file
            filesystem: PyArrow filesystem to use for accessing the file
            record_count: Number of records in the file
            source_content_length: Optional original content length in-memory
                before writing to disk.
            content_type: Optional content type override. If not provided, will
                be derived from file extension.
            content_encoding: Optional content encoding override. If not
                provided, will be derived from file extension.
            credentials: Optional credentials required to read this manifest entry.
            content_type_parameters: Optional content type parameters.
            entry_type: Optional entry type of this manifest entry. Defaults to DATA.
            entry_params: Optional entry type parameters.
            schema_id: Schema ID used to write this manifest entry.
            sort_scheme_id: Sort scheme ID used to write this manifest entry.

        Returns:
            A ManifestEntry instance
        """
        file_info = get_file_info(path, filesystem)
        if file_info.type != pa.fs.FileType.File:
            raise FileNotFoundError(f"Path does not point to a file: {path}")

        # Extract extensions from right to left
        # First split will get potential encoding extension
        base_path, ext1 = posixpath.splitext(path)

        # Initialize with defaults for no extensions
        derived_content_type = ContentType.BINARY
        derived_content_encoding = ContentEncoding.IDENTITY

        # Only proceed with extension checks if we found at least one extension
        if ext1:
            # Check if the first extension is a known encoding
            derived_content_encoding = EXT_TO_CONTENT_ENCODING.get(
                ext1,
                ContentEncoding.IDENTITY,
            )

            # Get second extension only if first was an encoding
            if derived_content_encoding != ContentEncoding.IDENTITY:
                # Second split will get potential content type extension
                _, ext2 = posixpath.splitext(base_path)
                if ext2:
                    derived_content_type = EXT_TO_CONTENT_TYPE.get(
                        ext2,
                        ContentType.BINARY,
                    )
            else:
                # First extension wasn't an encoding, check if it's a
                # content type
                derived_content_type = EXT_TO_CONTENT_TYPE.get(
                    ext1,
                    ContentType.BINARY,
                )

        if (
            derived_content_type == ContentType.BINARY
            and derived_content_encoding != ContentEncoding.IDENTITY
        ):
            logger.debug(
                f"Found encoding {derived_content_encoding.value} but no "
                f"content type for {path}, assuming binary"
            )

        # Use provided values if available, otherwise use derived values
        final_content_type = (
            content_type if content_type is not None else derived_content_type.value
        )
        final_content_encoding = (
            content_encoding
            if content_encoding is not None
            else derived_content_encoding.value
        )

        manifest_entry_meta = ManifestMeta.of(
            record_count=record_count,
            content_length=file_info.size,
            content_type=final_content_type,
            content_encoding=final_content_encoding,
            source_content_length=source_content_length,
            credentials=credentials,
            content_type_parameters=content_type_parameters,
            entry_type=entry_type,
            entry_params=entry_params,
            schema_id=schema_id,
            sort_scheme_id=sort_scheme_id,
        )
        manifest_entry = ManifestEntry.of(path, manifest_entry_meta)
        return manifest_entry

    @property
    def uri(self) -> Optional[str]:
        return self.get("uri")

    @property
    def url(self) -> Optional[str]:
        return self.get("url")

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def mandatory(self) -> bool:
        return self["mandatory"]

    @property
    def id(self) -> Optional[str]:
        return self.get("id")


class ManifestAuthor(dict):
    @staticmethod
    def of(name: Optional[str], version: Optional[str]) -> ManifestAuthor:
        manifest_author = ManifestAuthor()
        if name is not None:
            manifest_author["name"] = name
        if version is not None:
            manifest_author["version"] = version
        return manifest_author

    @staticmethod
    def from_dict(obj: dict) -> Optional[ManifestAuthor]:
        if obj is None:
            return None
        return ManifestAuthor.of(obj.get("name"), obj.get("version"))

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def version(self) -> Optional[str]:
        return self.get("version")


class ManifestEntryList(List[ManifestEntry]):
    @staticmethod
    def of(entries: List[ManifestEntry]) -> ManifestEntryList:
        manifest_entries = ManifestEntryList()
        for entry in entries:
            if entry is not None and not isinstance(entry, ManifestEntry):
                entry = ManifestEntry(entry)
            manifest_entries.append(entry)
        return manifest_entries

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, ManifestEntry):
            self[item] = val = ManifestEntry(val)
        return val

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]  # This triggers __getitem__ conversion



import argparse
import pathlib

from deltacat.compute import (
    job_client,
    JobStatus,
)


def run_async(
    source: str,
    dest: str,
    jobs_to_submit: int,
    job_timeout: int,
    cloud: str,
    restart_ray: bool,
):
    # print package version info
    working_dir = pathlib.Path(__file__).parent
    cluster_cfg_file_path = working_dir.joinpath(cloud).joinpath("deltacat.yaml")
    job_number = 0
    client = job_client(cluster_cfg_file_path, restart_ray=restart_ray)
    job_ids = []
    while jobs_to_submit > 0:
        jobs_to_submit -= 1
        job_dest = dest + f".{job_number}"
        job_id = client.submit_job(
            # Entrypoint shell command to execute
            entrypoint=f"python3 indexer.py --source '{source}' --dest '{job_dest}'",
            # Path to the local directory that contains the indexer.py file
            # This entire directory will be zipped into a job package, so keep
            # it small.
            runtime_env={"working_dir": working_dir},
        )
        job_ids.append(job_id)
        job_number += 1

    print("Waiting for all jobs to complete...")
    job_number = 0
    all_job_logs = ""
    for job_id in job_ids:
        job_status = client.await_job(job_id, timeout_seconds=job_timeout)
        if job_status != JobStatus.SUCCEEDED:
            print(f"Job `{job_id}` logs: ")
            print(client.get_job_logs(job_id))
            raise RuntimeError(f"Job `{job_id}` terminated with status: {job_status}")
        all_job_logs += f"\nJob #{job_number} logs: \n"
        all_job_logs += client.get_job_logs(job_id)
        job_number += 1
    print("All jobs completed!")
    print("Job Logs: ")
    print(all_job_logs)


def run_sync(
    source: str,
    dest: str,
    jobs_to_submit: int,
    job_timeout: int,
    cloud: str,
    restart_ray: bool,
):
    working_dir = pathlib.Path(__file__).parent
    cluster_cfg_file_path = working_dir.joinpath(cloud).joinpath("deltacat.yaml")
    client = job_client(cluster_cfg_file_path, restart_ray=restart_ray)
    job_number = 0
    while job_number < jobs_to_submit:
        job_dest = dest + f".{job_number}"
        job_run_result = client.run_job(
            # Entrypoint shell command to execute
            entrypoint=f"python3 indexer.py --source '{source}' --dest '{job_dest}'",
            # Path to the local directory that contains the indexer.py file
            # This entire directory will be zipped into a job package, so keep
            # it small.
            runtime_env={"working_dir": working_dir},
            timeout_seconds=job_timeout,
        )
        print(
            f"Job ID {job_run_result.job_id} terminal state: {job_run_result.job_status}"
        )
        print(f"Job ID {job_run_result.job_id} logs: ")
        print(job_run_result.job_logs)
        job_number += 1


def run(
    source: str,
    dest: str,
    restart_ray: bool,
    jobs_to_submit: int,
    job_timeout: int,
    asynchronous: bool,
    cloud_provider: str,
):
    run_func = run_async if asynchronous else run_sync
    run_func(
        source=source,
        dest=dest,
        jobs_to_submit=jobs_to_submit,
        job_timeout=job_timeout,
        cloud=cloud_provider,
        restart_ray=restart_ray,
    )


if __name__ == "__main__":
    """
    This example shows how to submit jobs to a remote Ray cluster that indexes source files into arbitrary destinations with
    optional file format conversion using DeltaCAT URLs. It provides the option to run multiple sequential or concurrent jobs
    for benchmarking.

    # For example, the following command launches a remote Ray Cluster on AWS, downloads an external OpenAlex dataset text file,
    # converts it to Parquet, and writes it back to AWS S3. It submits 100 jobs in parallel, each with a timeout of 90 seconds:
    $ python ./deltacat/examples/job_runner.py -- \
    $ --source text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31 \
    $ --dest parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet \
    $ --asynchronous \
    $ --jobs-to-submit 100 \
    $ --job-timeout 90 \
    $ --cloud-provider aws
    """
    script_args = [
        (
            [
                "--source",
            ],
            {
                "help": "Source DeltaCAT URL to index.",
                "type": str,
                "default": "text+s3://openalex-mag-format/data_dump_v1/2022-07-08/nlp/PaperAbstractsInvertedIndex.txt_part31",
            },
        ),
        (
            [
                "--dest",
            ],
            {
                "help": "Destination DeltaCAT URL to store the indexed file.",
                "type": str,
                "default": "parquet+s3://deltacat-example-output/openalex/PaperAbstractsInvertedIndex.part31.parquet",
            },
        ),
        (
            [
                "--restart-ray",
            ],
            {
                "help": "Restart Ray on an existing cluster.",
                "action": "store_true",
                "default": False,
            },
        ),
        (
            [
                "--asynchronous",
            ],
            {
                "help": "Run jobs asynchronously.",
                "action": "store_true",
                "default": False,
            },
        ),
        (
            [
                "--jobs-to-submit",
            ],
            {
                "help": "Number of indexer jobs to submit for execution.",
                "type": int,
                "default": 1,
            },
        ),
        (
            [
                "--job-timeout",
            ],
            {
                "help": "Job timeout in seconds.",
                "type": int,
                "default": 300,
            },
        ),
        (
            [
                "--cloud-provider",
            ],
            {
                "help": "Ray Cluster Cloud Provider ('aws' or 'gcp')",
                "type": str,
                "default": "aws",
            },
        ),
    ]

    # parse CLI input arguments
    parser = argparse.ArgumentParser()
    for args, kwargs in script_args:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"Command Line Arguments: {args}")

    # run the example using os.environ as kwargs
    run(**vars(args))



import unittest
from deltacat.exceptions import categorize_errors
import ray
from deltacat.exceptions import (
    DependencyPyarrowCapacityError,
    NonRetryableDownloadTableError,
    RetryableError,
    NonRetryableError,
    DeltaCatTransientError,
    DependencyDaftTransientError,
    UnclassifiedDeltaCatError,
)
from daft.exceptions import DaftTransientError
from deltacat.tests.utils.exceptions import (
    InvalidNamespaceError,
    MainStorageValidationError,
)
from deltacat.tests.utils import main_deltacat_storage_mock as ds
from botocore.exceptions import NoCredentialsError
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from pyarrow.lib import ArrowCapacityError


class MockUnknownException(Exception):
    pass


@categorize_errors
def mock_raise_exception(exception_to_raise, deltacat_storage=ds):
    raise exception_to_raise


@retry(retry=retry_if_exception_type(NoCredentialsError), stop=stop_after_attempt(2))
def mock_tenacity_wrapped_method(exception_to_raise):
    mock_raise_exception(exception_to_raise)


@ray.remote
def mock_remote_task(exception_to_raise):
    mock_raise_exception(exception_to_raise)


class TestCategorizeErrorsMain(unittest.TestCase):
    def test_pyarrow_exception_categorizer(self):
        self.assertRaises(
            DependencyPyarrowCapacityError,
            lambda: mock_raise_exception(ArrowCapacityError),
        )

    def test_storage_exception_categorizer(self):
        self.assertRaises(
            MainStorageValidationError,
            lambda: mock_raise_exception(InvalidNamespaceError, deltacat_storage=ds),
        )

    def test_non_retryable_error(self):
        self.assertRaises(
            NonRetryableError,
            lambda: mock_raise_exception(NonRetryableDownloadTableError),
        )

    def test_retryable_error(self):
        self.assertRaises(RetryableError, lambda: mock_raise_exception(ConnectionError))

    def test_ray_task_returns_wrapped_exception(self):
        self.assertRaises(
            DeltaCatTransientError,
            lambda: ray.get(mock_remote_task.remote(ConnectionError)),
        )

    def test_daft_transient_error(self):
        self.assertRaises(
            DependencyDaftTransientError,
            lambda: ray.get(mock_remote_task.remote(DaftTransientError)),
        )

    def test_tenacity_underlying_error_returned(self):
        self.assertRaises(
            DeltaCatTransientError,
            lambda: mock_tenacity_wrapped_method(NoCredentialsError),
        )

    def test_unclassified_error_when_error_cannot_be_categorized(self):
        self.assertRaises(
            UnclassifiedDeltaCatError,
            lambda: ray.get(mock_remote_task.remote(MockUnknownException)),
        )

    def test_deltacat_exception_contains_attributes(self):

        try:
            mock_raise_exception(ConnectionError)
        except DeltaCatTransientError as e:
            self.assertTrue(hasattr(e, "is_retryable"))
            self.assertTrue(hasattr(e, "error_name"))
            assert e.error_name == "DeltaCatTransientError"
            return

        self.assertFalse(True)


if __name__ == "__main__":
    unittest.main()



# TODO later on this will be moved to a dedicated package
from deltacat.experimental.storage.rivulet.parquet.file_reader import ParquetFileReader
from deltacat.experimental.storage.rivulet.reader.reader_type_registrar import (
    FileReaderRegistrar,
)

FileReaderRegistrar.register_reader("parquet", ParquetFileReader)



import logging
from typing import Optional
from deltacat import logs
from pyarrow.parquet import ColumnChunkMetaData
from deltacat.constants import NULL_SIZE_BYTES

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _observed_string_size(min_value: str, max_value: str) -> float:
    """
    Pyarrow uses few additional bytes to store each string.
    """
    return (len(min_value) + len(max_value)) / 2 + 4


def _int96_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values * 12


def _int64_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values * 8


def _int32_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values * 4


def _boolean_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values


def _double_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values * 8


def _float_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return column_chunk_metadata.num_values * 4


def _byte_array_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    uncompressed_size = column_chunk_metadata.total_uncompressed_size
    if column_chunk_metadata.is_stats_set:
        statistics = column_chunk_metadata.statistics
        if (
            statistics.has_min_max
            and isinstance(statistics.min, str)
            and isinstance(statistics.max, str)
        ):
            return max(
                uncompressed_size,
                (
                    statistics.num_values
                    * _observed_string_size(statistics.min, statistics.max)
                    + statistics.null_count * NULL_SIZE_BYTES
                ),
            )
        else:
            # A case of decimal
            return max(column_chunk_metadata.num_values * 16, uncompressed_size)
    else:
        return uncompressed_size


def _fixed_len_byte_array_size_estimator(
    column_chunk_metadata: ColumnChunkMetaData,
) -> float:
    return _byte_array_size_estimator(column_chunk_metadata)


_PHYSICAL_TYPE_TO_SIZE_ESTIMATOR = {
    "INT96": _int96_size_estimator,
    "INT64": _int64_size_estimator,
    "INT32": _int32_size_estimator,
    "BOOLEAN": _boolean_size_estimator,
    "DOUBLE": _double_size_estimator,
    "FLOAT": _float_size_estimator,
    "BYTE_ARRAY": _byte_array_size_estimator,
    "FIXED_LEN_BYTE_ARRAY": _fixed_len_byte_array_size_estimator,
}


def parquet_column_chunk_size_estimator(
    column_meta: ColumnChunkMetaData,
) -> Optional[float]:
    physical_type = column_meta.physical_type
    if physical_type in _PHYSICAL_TYPE_TO_SIZE_ESTIMATOR:
        return _PHYSICAL_TYPE_TO_SIZE_ESTIMATOR[physical_type](column_meta)
    else:
        logger.warning(
            f"Unsupported physical type: {physical_type}. "
            "Returning total_uncompressed_size."
        )
        return column_meta.total_uncompressed_size



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import base64
import re
import posixpath
from typing import Any, Dict, List, Optional, Tuple

import pyarrow
import pyarrow as pa

import deltacat.storage.model.partition as partition

from deltacat.storage.model.metafile import Metafile, MetafileRevisionInfo
from deltacat.constants import (
    METAFILE_FORMAT,
    METAFILE_FORMAT_JSON,
    TXN_DIR_NAME,
    BYTES_PER_KIBIBYTE,
)
from deltacat.storage.model.schema import (
    Schema,
    SchemaList,
)
from deltacat.storage.model.locator import (
    Locator,
    LocatorName,
)
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.table import (
    TableLocator,
    Table,
)
from deltacat.types.media import ContentType
from deltacat.storage.model.sort_key import SortScheme, SortSchemeList
from deltacat.storage.model.types import LifecycleState
from deltacat.types.tables import TableProperty

TableVersionProperties = Dict[str, Any]


class TableVersion(Metafile):
    @staticmethod
    def of(
        locator: Optional[TableVersionLocator],
        schema: Optional[Schema],
        partition_scheme: Optional[partition.PartitionScheme] = None,
        description: Optional[str] = None,
        properties: Optional[TableVersionProperties] = None,
        content_types: Optional[List[ContentType]] = None,
        sort_scheme: Optional[SortScheme] = None,
        watermark: Optional[int] = None,
        lifecycle_state: Optional[LifecycleState] = None,
        schemas: Optional[SchemaList] = None,
        partition_schemes: Optional[partition.PartitionSchemeList] = None,
        sort_schemes: Optional[SortSchemeList] = None,
        previous_table_version: Optional[str] = None,
        native_object: Optional[Any] = None,
    ) -> TableVersion:
        table_version = TableVersion()
        table_version.locator = locator
        table_version.schema = schema
        table_version.partition_scheme = partition_scheme
        table_version.description = description
        table_version.properties = properties
        table_version.content_types = content_types
        table_version.sort_scheme = sort_scheme
        table_version.watermark = watermark
        table_version.state = lifecycle_state
        table_version.schemas = schemas
        table_version.partition_schemes = partition_schemes
        table_version.sort_schemes = sort_schemes
        table_version.previous_table_version = previous_table_version
        table_version.native_object = native_object
        return table_version

    @property
    def locator(self) -> Optional[TableVersionLocator]:
        val: Dict[str, Any] = self.get("tableVersionLocator")
        if val is not None and not isinstance(val, TableVersionLocator):
            self.locator = val = TableVersionLocator(val)
        return val

    @locator.setter
    def locator(self, table_version_locator: Optional[TableVersionLocator]) -> None:
        self["tableVersionLocator"] = table_version_locator

    @property
    def schema(self) -> Optional[Schema]:
        val: Dict[str, Any] = self.get("schema")
        if val is not None and not isinstance(val, Schema):
            self.schema = val = Schema(val)
        return val

    @schema.setter
    def schema(self, schema: Optional[Schema]) -> None:
        self["schema"] = schema

    @property
    def schemas(self) -> Optional[SchemaList]:
        val: Optional[SchemaList] = self.get("schemas")
        if val is not None and not isinstance(val, SchemaList):
            self["schemas"] = val = SchemaList.of(val)
        return val

    @schemas.setter
    def schemas(self, schemas: Optional[SchemaList]) -> None:
        self["schemas"] = schemas

    @property
    def sort_scheme(self) -> Optional[SortScheme]:
        val: Dict[str, Any] = self.get("sortScheme")
        if val is not None and not isinstance(val, SortScheme):
            self["sortScheme"] = val = SortScheme(val)
        return val

    @sort_scheme.setter
    def sort_scheme(self, sort_scheme: Optional[SortScheme]) -> None:
        self["sortScheme"] = sort_scheme

    @property
    def sort_schemes(self) -> Optional[SortSchemeList]:
        val: Dict[str, Any] = self.get("sortSchemes")
        if val is not None and not isinstance(val, SortSchemeList):
            self["sortSchemes"] = val = SortSchemeList.of(val)
        return val

    @sort_schemes.setter
    def sort_schemes(self, sort_schemes: Optional[SortSchemeList]) -> None:
        self["sortSchemes"] = sort_schemes

    @property
    def watermark(self) -> Optional[int]:
        return self.get("watermark")

    @watermark.setter
    def watermark(self, watermark: Optional[int]) -> None:
        self["watermark"] = watermark

    @property
    def state(self) -> Optional[LifecycleState]:
        state = self.get("state")
        return None if state is None else LifecycleState(state)

    @state.setter
    def state(self, state: Optional[LifecycleState]) -> None:
        self["state"] = state

    @property
    def partition_scheme(self) -> Optional[partition.PartitionScheme]:
        val: Dict[str, Any] = self.get("partitionScheme")
        if val is not None and not isinstance(val, partition.PartitionScheme):
            self["partitionScheme"] = val = partition.PartitionScheme(val)
        return val

    @partition_scheme.setter
    def partition_scheme(
        self, partition_scheme: Optional[partition.PartitionScheme]
    ) -> None:
        self["partitionScheme"] = partition_scheme

    @property
    def partition_schemes(self) -> Optional[partition.PartitionSchemeList]:
        val: Dict[str, Any] = self.get("partitionSchemes")
        if val is not None and not isinstance(val, partition.PartitionSchemeList):
            self["partitionSchemes"] = val = partition.PartitionSchemeList.of(val)
        return val

    @partition_schemes.setter
    def partition_schemes(
        self, partition_schemes: Optional[partition.PartitionSchemeList]
    ) -> None:
        self["partitionSchemes"] = partition_schemes

    @property
    def description(self) -> Optional[str]:
        return self.get("description")

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self["description"] = description

    @property
    def previous_table_version(self) -> Optional[str]:
        return self.get("previous_table_version")

    @previous_table_version.setter
    def previous_table_version(self, previous_table_version: Optional[str]) -> None:
        self["previous_table_version"] = previous_table_version

    @property
    def properties(self) -> Optional[TableVersionProperties]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[TableVersionProperties]) -> None:
        self["properties"] = properties

    @property
    def content_types(self) -> Optional[List[ContentType]]:
        content_types = self.get("contentTypes")
        return (
            None
            if content_types is None
            else [None if _ is None else ContentType(_) for _ in content_types]
        )

    @content_types.setter
    def content_types(self, content_types: Optional[List[ContentType]]) -> None:
        self["contentTypes"] = content_types

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")

    @native_object.setter
    def native_object(self, native_object: Optional[Any]) -> None:
        self["nativeObject"] = native_object

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_version
        return None

    def url(self, catalog_name: Optional[str] = None) -> str:
        return (
            f"dc://{catalog_name}/{self.namespace}/{self.table_name}/{self.table_version}/"
            if catalog_name
            else f"table://{self.namespace}/{self.table_name}/{self.table_version}/"
        )

    def is_supported_content_type(self, content_type: ContentType):
        supported_content_types = self.content_types
        return (not supported_content_types) or (
            content_type in supported_content_types
        )

    def to_serializable(self) -> TableVersion:
        serializable: TableVersion = TableVersion.update_for(self)
        if serializable.schema:
            schema_bytes = serializable.schema.serialize().to_pybytes()
            serializable.schema = (
                base64.b64encode(schema_bytes).decode("utf-8")
                if METAFILE_FORMAT == METAFILE_FORMAT_JSON
                else schema_bytes
            )

        if serializable.schemas:
            serializable.schemas = [
                base64.b64encode(schema.serialize().to_pybytes()).decode("utf-8")
                if METAFILE_FORMAT == METAFILE_FORMAT_JSON
                else schema.serialize().to_pybytes()
                for schema in serializable.schemas
            ]
        if serializable.table_locator:
            # remove the mutable table locator
            serializable.locator.table_locator = TableLocator.at(
                namespace=self.id,
                table_name=self.id,
            )
        return serializable

    def from_serializable(
        self,
        path: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> TableVersion:
        if self.get("schema"):
            schema_data = self["schema"]
            schema_bytes = (
                base64.b64decode(schema_data)
                if METAFILE_FORMAT == "json"
                else schema_data
            )
            self["schema"] = Schema.deserialize(pa.py_buffer(schema_bytes))
        else:
            self["schema"] = None

        if self.get("schemas"):
            self.schemas = [
                Schema.deserialize(
                    pa.py_buffer(
                        base64.b64decode(schema)
                        if METAFILE_FORMAT == METAFILE_FORMAT_JSON
                        else schema
                    )
                )
                for schema in self["schemas"]
            ]
        else:
            self.schemas = None

        if self.sort_scheme:
            # force list-to-tuple conversion of sort keys via property invocation
            self.sort_scheme.keys
            [sort_scheme.keys for sort_scheme in self.sort_schemes]
        # restore the table locator from its mapped immutable metafile ID
        if self.table_locator and self.table_locator.table_name == self.id:
            parent_rev_dir_path = Metafile._parent_metafile_rev_dir_path(
                base_metafile_path=path,
                parent_number=1,
            )
            txn_log_dir = posixpath.join(
                posixpath.dirname(
                    posixpath.dirname(
                        posixpath.dirname(parent_rev_dir_path),
                    )
                ),
                TXN_DIR_NAME,
            )
            table = Table.read(
                MetafileRevisionInfo.latest_revision(
                    revision_dir_path=parent_rev_dir_path,
                    filesystem=filesystem,
                    success_txn_log_dir=txn_log_dir,
                ).path,
                filesystem,
            )
            self.locator.table_locator = table.locator
        return self

    def current_version_number(self) -> Optional[int]:
        """
        Returns the current table version number as an integer, or None if
        a table version has not yet been assigned.
        """
        prefix, version_number = (
            TableVersion.parse_table_version(
                self.table_version,
            )
            if self.table_version is not None
            else (None, None)
        )
        return int(version_number) if version_number is not None else None

    def read_table_property(self, property: TableProperty) -> Any:
        return TableProperty.read_table_property(self, property)

    @staticmethod
    def next_version(previous_version: Optional[str] = None) -> str:
        """
        Assigns the next table version string given the previous table version
        by incrementing the version number of the given previous table version
        identifier. Returns "1" if the previous version is undefined.
        """
        prefix, previous_version_number = (
            TableVersion.parse_table_version(
                previous_version,
            )
            if previous_version is not None
            else (None, None)
        )
        new_version_number = (
            int(previous_version_number) + 1
            if previous_version_number is not None
            else 1
        )
        new_prefix = prefix if prefix is not None else ""
        return f"{new_prefix}{new_version_number}"

    @staticmethod
    def parse_table_version(table_version: str) -> Tuple[Optional[str], int]:
        """
        Parses a table version string into its prefix and version number.
        Returns a tuple of the prefix and version number.
        """
        if not table_version:
            raise ValueError(f"Table version to parse is undefined.")
        if len(table_version) > BYTES_PER_KIBIBYTE:
            raise ValueError(
                f"Invalid table version {table_version}. Table version "
                f"identifier cannot be greater than {BYTES_PER_KIBIBYTE} "
                f"characters."
            )
        version_match = re.match(
            rf"^(\w*\.)?(\d+)$",
            table_version,
        )
        if version_match:
            prefix, version_number = version_match.groups()
            return prefix, int(version_number)
        raise ValueError(
            f"Invalid table version {table_version}. Valid table versions "
            f"are of the form `TableVersionName.1` or simply `1`.",
        )


class TableVersionLocatorName(LocatorName):
    def __init__(self, locator: TableVersionLocator):
        self.locator = locator

    @property
    def immutable_id(self) -> Optional[str]:
        return self.locator.table_version

    @immutable_id.setter
    def immutable_id(self, immutable_id: Optional[str]):
        self.locator.table_version = immutable_id

    def parts(self) -> List[str]:
        return [self.locator.table_version]


class TableVersionLocator(Locator, dict):
    @staticmethod
    def of(
        table_locator: Optional[TableLocator],
        table_version: Optional[str],
    ) -> TableVersionLocator:
        table_version_locator = TableVersionLocator()
        table_version_locator.table_locator = table_locator
        table_version_locator.table_version = table_version
        return table_version_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
    ) -> TableVersionLocator:
        table_locator = TableLocator.at(namespace, table_name) if table_name else None
        return TableVersionLocator.of(table_locator, table_version)

    @property
    def name(self):
        return TableVersionLocatorName(self)

    @property
    def parent(self) -> Optional[TableLocator]:
        return self.table_locator

    @property
    def table_locator(self) -> Optional[TableLocator]:
        val: Dict[str, Any] = self.get("tableLocator")
        if val is not None and not isinstance(val, TableLocator):
            self.table_locator = val = TableLocator(val)
        return val

    @table_locator.setter
    def table_locator(self, table_locator: Optional[TableLocator]) -> None:
        self["tableLocator"] = table_locator

    @property
    def table_version(self) -> Optional[str]:
        return self.get("tableVersion")

    @table_version.setter
    def table_version(self, table_version: Optional[str]) -> None:
        # ensure that the table version is valid
        prefix, version_number = TableVersion.parse_table_version(table_version)
        # restate the table version number in its canonical form
        # (e.g., ensure that "MyVersion.0001" is saved as "MyVersion.1")
        self["tableVersion"] = (
            f"{prefix}{version_number}" if prefix else str(version_number)
        )

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.namespace_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.table_name
        return None



from unittest import TestCase

import ray.util.state
from deltacat.utils.ray_utils.concurrency import invoke_parallel
import ray
import time


@ray.remote
def f(x):
    return time.sleep(x)


class TestInvokeParallel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        ray.init(local_mode=True, ignore_reinit_error=True)

    def test_invoke_parallel_when_max_parallelism_is_list_size(self):

        items = [1 for i in range(3)]

        result = invoke_parallel(items=items, ray_task=f, max_parallelism=3)

        self.assertEqual(3, len(result))

    def test_invoke_parallel_when_max_parallelism_is_less_than_list_size(self):
        @ray.remote
        def f(x):
            return x

        items = [i for i in range(3)]

        result = invoke_parallel(items=items, ray_task=f, max_parallelism=1)

        self.assertEqual(3, len(result))

    def test_invoke_parallel_with_options_provider(self):

        items = [1 for i in range(1)]

        result = invoke_parallel(
            items=items,
            ray_task=f,
            max_parallelism=10,
            options_provider=lambda x, y: {"memory": 1},
        )

        self.assertEqual(1, len(result))



from typing import Optional, Dict, Any
from pyiceberg.table import Table
from pyiceberg.catalog import Catalog, load_catalog as pyiceberg_load_catalog
from botocore.credentials import Credentials
import boto3
from boto3.session import Session


def load_catalog(
    iceberg_catalog_name: str, iceberg_catalog_properties: Dict[str, Any]
) -> Catalog:
    catalog = pyiceberg_load_catalog(
        name=iceberg_catalog_name,
        **iceberg_catalog_properties,
    )
    return catalog


def get_s3_path(
    bucket_name: str,
    database_name: Optional[str] = None,
    table_name: Optional[str] = None,
) -> str:
    result_path = f"s3://{bucket_name}"
    if database_name is not None:
        result_path += f"/{database_name}.db"

    if table_name is not None:
        result_path += f"/{table_name}"
    return result_path


def get_bucket_name() -> str:
    return "test-bucket"


def get_s3_prefix() -> str:
    return get_s3_path(get_bucket_name())


def get_credential() -> Credentials:
    boto3_session: Session = boto3.Session()
    credentials: Credentials = boto3_session.get_credentials()
    return credentials


def get_glue_catalog() -> Catalog:
    credential = get_credential()
    # Credentials are refreshable, so accessing your access key / secret key
    # separately can lead to a race condition. Use this to get an actual matched
    # set.
    credential = credential.get_frozen_credentials()
    access_key_id = credential.access_key
    secret_access_key = credential.secret_key
    session_token = credential.token
    s3_path = get_s3_prefix()
    glue_catalog = pyiceberg_load_catalog(
        "glue",
        **{
            "warehouse": s3_path,
            "type": "glue",
            "aws_access_key_id": access_key_id,
            "aws_secret_access_key": secret_access_key,
            "aws_session_token": session_token,
            "region_name": "us-east-1",
            "s3.access-key-id": access_key_id,
            "s3.secret-access-key": secret_access_key,
            "s3.session-token": session_token,
            "s3.region": "us-east-1",
        },
    )

    return glue_catalog


def load_table(catalog: Catalog, table_name: str) -> Table:
    loaded_table = catalog.load_table(table_name)
    return loaded_table
